"""Classes for modeling a DAG."""

from __future__ import annotations

import html
import logging
import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, List, Sequence, TypeVar, Generic

import jinja2

from metricflow.dag.id_generation import IdGeneratorRegistry
from metricflow.visitor import VisitorOutputT

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class DisplayedProperty:  # type: ignore
    """When visualizing a node as text or a graphic, the properties to display with it.

    This should be displayed in the form "{key} = {str(value)}"
    """

    key: str
    value: Any  # type: ignore


@dataclass(frozen=True)
class NodeId:
    """Unique identifier for nodes in DAGs."""

    id_str: str

    def __repr__(self) -> str:  # noqa: D
        return self.id_str


class DagNodeVisitor(Generic[VisitorOutputT], ABC):
    """An object that can be used to visit the nodes of a DAG graph"""

    @abstractmethod
    def visit_node(self, node: DagNode) -> VisitorOutputT:  # noqa: D
        pass


class DagNode(ABC):
    """A node in a DAG. These should be immutable."""

    def __init__(self, node_id: NodeId) -> None:  # noqa: D
        self._node_id = node_id

    @property
    def node_id(self) -> NodeId:
        """ID for uniquely identifying a given node."""
        return self._node_id

    @property
    @abstractmethod
    def description(self) -> str:
        """A human-readable description for this node."""
        pass

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:
        """When the node is represented in a visualization, the properties to display with it."""
        return [
            DisplayedProperty("description", self.description),
            DisplayedProperty("node_id", self.node_id),
        ]

    @property
    def graphviz_label(self) -> str:
        """When using graphviz to render this node, the label that should be used in the construction."""
        return make_graphviz_label(
            title=self.__class__.__name__,
            properties=self.displayed_properties,
        )

    @property
    @abstractmethod
    def parent_nodes(self) -> Sequence[DagNode]:  # noqa: D
        pass

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}(node_id={self.node_id})"

    @classmethod
    @abstractmethod
    def id_prefix(cls) -> str:
        """The prefix to use when generating IDs for nodes.

        e.g. a prefix of "my_node" will generate an ID like "my_node_0"
        """
        pass

    @classmethod
    def create_unique_id(cls) -> NodeId:
        """Create and return a unique identifier to use when creating nodes."""
        return NodeId(IdGeneratorRegistry.for_class(cls).create_id(cls.id_prefix()))

    def accept_dag_node_visitor(self, visitor: DagNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        """Visit this node."""
        return visitor.visit_node(self)


def make_graphviz_label(
    title: str, properties: List[DisplayedProperty], title_font_size: int = 12, property_font_size: int = 6
) -> str:
    """Make a graphviz label that can be used for rendering to an image.

    The title will be in a large font, while the properties will be listed in a table in a smaller font.
    """
    # Convert all properties values into a HTML-safe string, then break the string into lines of 40 columns so that
    # the node boxes don't get so wide. Better to pretty-print the object, but unclear how to do so.
    formatted_properties = []
    for displayed_property in properties:
        lines = [html.escape(x) for x in textwrap.wrap(str(displayed_property.value), width=40)]
        formatted_properties.append(DisplayedProperty(displayed_property.key, "<BR/>".join(lines)))

    return jinja2.Template(
        # Formatting here: https://graphviz.org/doc/info/shapes.html#html
        textwrap.dedent(
            """\
            <<TABLE BORDER="0" CELLPADDING="1" CELLSPACING="0">
             <TR>
               <TD ALIGN="LEFT" BALIGN="LEFT" VALIGN="TOP" COLSPAN="2"><FONT point-size="{{ title_size }}">{{ title }}</FONT></TD>
             </TR>
             {%- for key, value in properties %}
             <TR>
               <TD ALIGN="LEFT" BALIGN="LEFT" VALIGN="TOP"><FONT point-size="{{ property_size }}">{{ key }}</FONT></TD>
               <TD ALIGN="LEFT" BALIGN="LEFT" VALIGN="TOP"><FONT point-size="{{ property_size }}">{{ value }}</FONT></TD>
             </TR>
             {%- endfor %}
            </TABLE>>
            """
        ),
        undefined=jinja2.StrictUndefined,
    ).render(
        title=title,
        title_size=title_font_size,
        property_size=property_font_size,
        properties=[(displayed_property.key, displayed_property.value) for displayed_property in formatted_properties],
    )


DagNodeT = TypeVar("DagNodeT", bound=DagNode)


class MetricFlowDag(Generic[DagNodeT]):  # noqa: D
    """Represents a directed acyclic graph. The sink nodes will have the connected components."""

    def __init__(self, dag_id: str, sink_nodes: List[DagNodeT]):  # noqa: D
        self._dag_id = dag_id
        self._sink_nodes = sink_nodes

    @property
    def dag_id(self) -> str:  # noqa: D
        return self._dag_id

    @property
    def sink_nodes(self) -> List[DagNodeT]:  # noqa: D
        return self._sink_nodes



import pytest

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.query.query_parser import MetricFlowQueryParser
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository


@pytest.fixture(scope="session")
def query_parser(  # noqa: D
    simple_semantic_model: SemanticModel,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    time_spine_source: TimeSpineSource,
) -> MetricFlowQueryParser:
    return MetricFlowQueryParser(
        model=simple_semantic_model,
        source_nodes=consistent_id_object_repository.simple_model_source_nodes,
        node_output_resolver=DataflowPlanNodeOutputDataSetResolver(
            column_association_resolver=DefaultColumnAssociationResolver(simple_semantic_model),
            semantic_model=simple_semantic_model,
            time_spine_source=time_spine_source,
        ),
    )



from typing import Sequence, List

from metricflow.plan_conversion.select_column_gen import SelectColumnSet
from metricflow.specs import (
    InstanceSpecSetTransform,
    InstanceSpecSet,
    ColumnAssociationResolver,
    TransformOutputT,
)
from metricflow.sql.sql_exprs import (
    SqlExpressionNode,
    SqlLogicalExpression,
    SqlLogicalOperator,
    SqlComparisonExpression,
    SqlComparison,
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlFunctionExpression,
    SqlFunction,
)
from metricflow.sql.sql_plan import SqlSelectColumn


def _make_coalesced_expr(table_aliases: Sequence[str], column_alias: str) -> SqlExpressionNode:
    """Makes a coalesced expression of the given column from the given table aliases.

    e.g.

    table_aliases = ["a", "b"]
    column_alias = "is_instant"

    ->

    COALESCE(a.is_instant, b.is_instant)
    """
    if len(table_aliases) == 1:
        return SqlColumnReferenceExpression(
            col_ref=SqlColumnReference(
                table_alias=table_aliases[0],
                column_name=column_alias,
            )
        )
    else:
        columns_to_coalesce: List[SqlExpressionNode] = []
        for table_alias in table_aliases:
            columns_to_coalesce.append(
                SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(
                        table_alias=table_alias,
                        column_name=column_alias,
                    )
                )
            )
        return SqlFunctionExpression(
            sql_function=SqlFunction.COALESCE,
            sql_function_args=columns_to_coalesce,
        )


class CreateOnConditionForCombiningMetrics(InstanceSpecSetTransform[SqlExpressionNode]):
    """Creates an expression that can go in the ON condition when coalescing linkables with a FULL OUTER JOIN.

    e.g.

    dimension_specs = ["is_instant", "ds"]
    table_aliases_in_coalesce = ["a", "b"]
    table_alias_on_right_equality = ["c"]

    ->

    COALESCE(a.is_instant, b.is_instant) = c.is_instant
    AND COALESCE(a.ds, b.ds) = c.ds

    """

    def __init__(  # noqa: D
        self,
        column_association_resolver: ColumnAssociationResolver,
        table_aliases_in_coalesce: Sequence[str],
        table_alias_on_right_equality: str,
    ) -> None:
        self._column_association_resolver = column_association_resolver
        self._table_aliases_in_coalesce = table_aliases_in_coalesce
        self._table_alias_on_right_equality = table_alias_on_right_equality

    def _make_equality_expr(self, column_alias: str) -> SqlExpressionNode:
        return SqlComparisonExpression(
            left_expr=_make_coalesced_expr(self._table_aliases_in_coalesce, column_alias),
            comparison=SqlComparison.EQUALS,
            right_expr=SqlColumnReferenceExpression(
                col_ref=SqlColumnReference(
                    table_alias=self._table_alias_on_right_equality,
                    column_name=column_alias,
                )
            ),
        )

    def transform(self, spec_set: InstanceSpecSet) -> SqlExpressionNode:  # noqa: D
        equality_exprs = []

        for dimension_spec in spec_set.dimension_specs:
            equality_exprs.append(
                self._make_equality_expr(
                    column_alias=self._column_association_resolver.resolve_dimension_spec(dimension_spec).column_name
                )
            )

        for time_dimension_spec in spec_set.time_dimension_specs:
            equality_exprs.append(
                self._make_equality_expr(
                    column_alias=self._column_association_resolver.resolve_time_dimension_spec(
                        time_dimension_spec
                    ).column_name
                )
            )

        for identifier_spec in spec_set.identifier_specs:
            column_associations = self._column_association_resolver.resolve_identifier_spec(identifier_spec)
            assert len(column_associations) == 1, "Composite identifiers not supported"
            column_association = column_associations[0]

            equality_exprs.append(self._make_equality_expr(column_alias=column_association.column_name))
        assert len(equality_exprs) > 0

        if len(equality_exprs) == 1:
            return equality_exprs[0]
        else:
            return SqlLogicalExpression(
                operator=SqlLogicalOperator.AND,
                args=tuple(equality_exprs),
            )


class CreateSelectCoalescedColumnsForLinkableSpecs(InstanceSpecSetTransform[SelectColumnSet]):
    """Create select columns that coalesce columns corresponding to linkable specs.

    e.g.

    dimension_specs = [DimensionSpec(element_name="is_instant")]
    table_aliases = ["a", "b"]

    ->

    COALESCE(a.is_instant, b.is_instant) AS is_instant
    """

    def __init__(  # noqa: D
        self,
        column_association_resolver: ColumnAssociationResolver,
        table_aliases: Sequence[str],
    ) -> None:
        self._column_association_resolver = column_association_resolver
        self._table_aliases = table_aliases

    def transform(self, spec_set: InstanceSpecSet) -> SelectColumnSet:  # noqa: D

        dimension_columns: List[SqlSelectColumn] = []
        time_dimension_columns: List[SqlSelectColumn] = []
        identifier_columns: List[SqlSelectColumn] = []

        for dimension_spec in spec_set.dimension_specs:
            column_name = self._column_association_resolver.resolve_dimension_spec(dimension_spec).column_name
            dimension_columns.append(
                SqlSelectColumn(
                    expr=_make_coalesced_expr(self._table_aliases, column_name),
                    column_alias=column_name,
                )
            )

        for time_dimension_spec in spec_set.time_dimension_specs:
            column_name = self._column_association_resolver.resolve_time_dimension_spec(time_dimension_spec).column_name
            time_dimension_columns.append(
                SqlSelectColumn(
                    expr=_make_coalesced_expr(self._table_aliases, column_name),
                    column_alias=column_name,
                )
            )

        for identifier_spec in spec_set.identifier_specs:
            column_associations = self._column_association_resolver.resolve_identifier_spec(identifier_spec)
            assert len(column_associations) == 1, "Composite identifiers not supported"
            column_name = column_associations[0].column_name

            identifier_columns.append(
                SqlSelectColumn(
                    expr=_make_coalesced_expr(self._table_aliases, column_name),
                    column_alias=column_name,
                )
            )

        return SelectColumnSet(
            dimension_columns=dimension_columns,
            time_dimension_columns=time_dimension_columns,
            identifier_columns=identifier_columns,
        )


class SelectOnlyLinkableSpecs(InstanceSpecSetTransform[InstanceSpecSet]):
    """Removes metrics and measures from the spec set."""

    def transform(self, spec_set: InstanceSpecSet) -> TransformOutputT:  # noqa: D
        return InstanceSpecSet(
            metric_specs=(),
            measure_specs=(),
            dimension_specs=spec_set.dimension_specs,
            time_dimension_specs=spec_set.time_dimension_specs,
            identifier_specs=spec_set.identifier_specs,
        )



from __future__ import annotations

import logging
import os
from collections import OrderedDict
from enum import Enum
from typing import Sequence, Tuple, Optional

import yaml

from metricflow.model.objects.base import FrozenBaseModel

logger = logging.getLogger(__name__)

DOCUMENT_KEY = "integration_test"


class IntegrationTestModel(Enum):
    """Names of the models that are possible to specify in the test case files."""

    SIMPLE_MODEL = "SIMPLE_MODEL"
    SIMPLE_MODEL_NON_DS = "SIMPLE_MODEL_NON_DS"
    COMPOSITE_IDENTIFIER_MODEL = "COMPOSITE_IDENTIFIER_MODEL"
    UNPARTITIONED_MULTI_HOP_JOIN_MODEL = "UNPARTITIONED_MULTI_HOP_JOIN_MODEL"
    PARTITIONED_MULTI_HOP_JOIN_MODEL = "PARTITIONED_MULTI_HOP_JOIN_MODEL"
    EXTENDED_DATE_MODEL = "EXTENDED_DATE_MODEL"


class RequiredDwEngineFeatures(Enum):
    """Required features that are needed in the DW engine for the test to run."""

    DATE_TRUNC = "DATE_TRUNC"
    FULL_OUTER_JOIN = "FULL_OUTER_JOIN"

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}.{self.name}"


class ConfiguredIntegrationTestCase(FrozenBaseModel):
    """Integration test case parsed from YAML files."""

    # Pydantic feature to throw errors on extra fields.
    class Config:  # noqa: D
        extra = "forbid"

    # Name of the test.
    name: str
    # Name of the semantic model to use.
    model: IntegrationTestModel
    metrics: Tuple[str, ...]
    # The SQL query that can be run to obtain the expected results.
    check_query: str
    file_path: str
    group_bys: Tuple[str, ...] = ()
    order_bys: Tuple[str, ...] = ()
    # The required features in the DW engine for the test to complete.
    required_features: Tuple[RequiredDwEngineFeatures, ...] = ()
    # Whether to check the order of the rows / columns.
    check_order: bool = False
    allow_empty: bool = False
    time_constraint: Optional[Tuple[str, str]] = None
    where_constraint: Optional[str] = None
    limit: Optional[int] = None
    description: Optional[str] = None


class TestCaseParseException(Exception):
    """Exception thrown when there is an error parsing the YAML test configuration."""

    pass


class ConfiguredIntegrationTestCaseRepository:
    """Stores integration test cases generated by parsing YAML files."""

    def __init__(self, config_directory: str) -> None:
        """Constructor.

        Args:
            config_directory: directory that should be searched for YAML files containing test cases.
        """
        self._config_directory = config_directory
        self._test_case_file_paths = ConfiguredIntegrationTestCaseRepository._find_all_yaml_file_paths(
            self._config_directory
        )
        self._test_cases: OrderedDict = OrderedDict()

        for file_path in self._test_case_file_paths:
            test_cases = ConfiguredIntegrationTestCaseRepository._parse_config_yaml(file_path)
            for test_case in test_cases:
                qualified_name = f"{os.path.basename(file_path)}/{test_case.name}"
                if qualified_name in self._test_cases:
                    raise ValueError(f"Test with a duplicate test name found: {test_case.name}")
                self._test_cases[qualified_name] = test_case

    @staticmethod
    def _parse_config_yaml(file_path: str) -> Sequence[ConfiguredIntegrationTestCase]:
        """Parse the YAML file at the given path into test cases."""
        results = []

        with open(file_path) as f:
            file_contents = f.read()
            for config_document in yaml.load_all(stream=file_contents, Loader=yaml.SafeLoader):
                # The config document can be None if there is nothing but white space between two `---`
                # this isn't really an issue, so lets just swallow it
                if config_document is None:
                    continue
                if not isinstance(config_document, dict):
                    raise TestCaseParseException(
                        f"Test query object YAML must be a dict. Got `{type(config_document)}`: {config_document}"
                    )

                keys = tuple(x for x in config_document.keys())
                if len(keys) != 1:
                    raise TestCaseParseException(
                        f"Test case document should have one type of key, but this has {len(keys)}. "
                        f"Found keys: {keys} in {file_path}",
                    )

                # retrieve last top-level key as type
                document_type = next(iter(config_document.keys()))
                object_cfg = config_document[document_type]
                if document_type == DOCUMENT_KEY:
                    try:
                        results.append(ConfiguredIntegrationTestCase(**object_cfg, file_path=file_path))
                    except Exception as e:
                        raise TestCaseParseException(f"Error while parsing: {file_path}") from e
                else:
                    raise TestCaseParseException(f"Expected {DOCUMENT_KEY}, but got {document_type}")
        return results

    @staticmethod
    def _find_all_yaml_file_paths(directory: str) -> Sequence[str]:  # noqa: D
        """Recursively search through the given directory for YAML files."""
        test_case_file_paths = []

        for root, dirs, files in os.walk(directory):
            for file in files:
                if file.endswith(".yaml"):
                    test_case_file_paths.append(os.path.join(root, file))

        return sorted(test_case_file_paths)

    @property
    def all_test_case_names(self) -> Sequence[str]:
        """Return test case names as specified in the YAML files."""
        return tuple(self._test_cases.keys())

    def get_test_case(self, test_case_name: str) -> ConfiguredIntegrationTestCase:
        """Get a specific test case by name. Throws an error if the given test case does not exist."""
        if test_case_name not in self._test_cases:
            raise ValueError(f"Unknown test case: {test_case_name}")
        return self._test_cases[test_case_name]


CONFIGURED_INTEGRATION_TESTS_REPOSITORY = ConfiguredIntegrationTestCaseRepository(
    os.path.join(os.path.dirname(__file__), "test_cases"),
)



import textwrap
from typing import Dict, Optional

from metricflow.model.parsing.yaml_loader import ParsingContext


class CustomerFacingSemanticException(Exception):
    """Class of Exceptions that make it to the customer's eyeballs"""

    pass


class UnableToSatisfyQueryError(CustomerFacingSemanticException):  # noqa:D
    def __init__(self, error_name: str, context: Dict[str, str] = None) -> None:  # noqa:D
        """Context will be printed as list of items when this is converted to a string."""
        self.error_name = error_name
        self._context = context

    def __str__(self) -> str:  # noqa:D
        error_lines = ["Unable To Satisfy Query Error: " + self.error_name]
        if self._context:
            for key, value in self._context.items():
                error_lines.append(f"\n{key}:")
                error_lines.append(f"{textwrap.indent(value, prefix='    ')}")
        return "\n".join(error_lines)


class SemanticException(Exception):  # noqa:D
    pass


class DuplicateMetricError(SemanticException):  # noqa:D
    pass


class MaterializationNotFoundError(SemanticException, KeyError):  # noqa:D
    pass


class MetricNotFoundError(SemanticException, KeyError):  # noqa:D
    pass


class NonExistentMeasureError(SemanticException):  # noqa:D
    pass


class InvalidDataSourceError(SemanticException):  # noqa:D
    pass


class ConstraintParseException(Exception):  # noqa: D
    pass


class ParsingException(Exception):  # noqa: D
    def __init__(  # noqa: D
        self, message: str, ctx: Optional[ParsingContext] = None, config_filepath: Optional[str] = None
    ) -> None:
        if config_filepath:
            message = f"Failed to parse YAML file '{config_filepath}' - {message}"
        if ctx:
            message = f"{message}\nContext: {str(ctx)}"
        super().__init__(message)


class ExecutionException(Exception):
    """Raised if there are any errors while executing the execution plan"""

    pass


class SqlClientCreationException(Exception):
    """Exception to represent errors related to the SqlClient"""

    def __init__(self) -> None:  # noqa: D
        super().__init__(
            "An error occurred when attempting to create the SqlClient",
        )


class MetricFlowInitException(Exception):
    """Exception to represent errors related to the MetricFlow creation"""

    def __init__(self) -> None:  # noqa: D
        super().__init__(
            "An error occurred when attempting to create the MetricFlow engine",
        )


class ModelCreationException(Exception):
    """Exception to represent errors related to the building a model"""

    def __init__(self) -> None:  # noqa: D
        super().__init__(
            "An error occurred when attempting to build the semantic model",
        )



import datetime
from metricflow.time.time_source import TimeSource


class ConfigurableTimeSource(TimeSource):
    """A time source that can be configured so that scheduled operations can be simulated in testing."""

    def __init__(self, configured_time: datetime.datetime) -> None:  # noqa: D
        self._configured_time = configured_time

    def get_time(self) -> datetime.datetime:  # noqa: D
        return self._configured_time

    def set_time(self, new_time: datetime.datetime) -> datetime.datetime:  # noqa: D
        self._configured_time = new_time
        return new_time



from __future__ import annotations

import collections
import logging
import time
from dataclasses import dataclass
from typing import List, TypeVar, Optional, Generic, Dict, Tuple, Sequence, Set, Union

from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.dag.id_generation import IdGeneratorRegistry, DATAFLOW_PLAN_PREFIX
from metricflow.dataflow.builder.costing import DefaultCostFunction, DataflowPlanNodeCostFunction
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.node_evaluator import (
    NodeEvaluatorForLinkableInstances,
    JoinLinkableInstancesRecipe,
    LinkableInstanceSatisfiabilityEvaluation,
)
from metricflow.dataflow.dataflow_plan import (
    AggregateMeasuresNode,
    BaseOutput,
    CombineMetricsNode,
    ComputeMetricsNode,
    ComputedMetricsOutput,
    ConstrainTimeRangeNode,
    DataflowPlan,
    FilterElementsNode,
    JoinAggregatedMeasuresByGroupByColumnsNode,
    JoinDescription,
    JoinOverTimeRangeNode,
    JoinToBaseOutputNode,
    OrderByLimitNode,
    WhereConstraintNode,
    WriteToResultDataframeNode,
    WriteToResultTableNode,
    SinkOutput,
)
from metricflow.dataflow.dataflow_plan_to_text import dataflow_dag_as_text
from metricflow.dataflow.sql_table import SqlTable
from metricflow.dataset.dataset import DataSet
from metricflow.errors.errors import UnableToSatisfyQueryError
from metricflow.model.objects.metric import MetricType, CumulativeMetricWindow
from metricflow.model.semantic_model import SemanticModel
from metricflow.object_utils import pformat_big_objects, assert_exactly_one_arg_set
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.node_processor import PreDimensionJoinNodeProcessor
from metricflow.plan_conversion.sql_dataset import SqlDataSet
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.query.query_parser import MetricFlowQueryParser
from metricflow.specs import (
    MetricSpec,
    LinkableInstanceSpec,
    MetricFlowQuerySpec,
    MeasureSpec,
    TimeDimensionSpec,
    IdentifierSpec,
    InstanceSpec,
    DimensionSpec,
    OrderBySpec,
    SpecWhereClauseConstraint,
    LinkableSpecSet,
    ColumnAssociationResolver,
)
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)

# The type of data set that at the source nodes.
SqlDataSetT = TypeVar("SqlDataSetT", bound=SqlDataSet)


@dataclass(frozen=True)
class MeasureRecipe(Generic[SqlDataSetT]):
    """Get a recipe for how to build a dataflow plan node that outputs measures and the needed linkable instances.

    The recipe involves filtering the measure node so that it only outputs the measures and the instances associated with
    required_local_linkable_specs, then joining the nodes containing the linkable instances according to the recipes
    in join_linkable_instances_recipes.
    """

    measure_node: BaseOutput[SqlDataSetT]
    required_local_linkable_specs: Tuple[LinkableInstanceSpec, ...]
    join_linkable_instances_recipes: Tuple[JoinLinkableInstancesRecipe, ...]


class DataflowPlanBuilder(Generic[SqlDataSetT]):
    """Builds a dataflow plan to satisfy a given query."""

    def __init__(  # noqa: D
        self,
        source_nodes: Sequence[BaseOutput[SqlDataSetT]],
        semantic_model: SemanticModel,
        time_spine_source: TimeSpineSource,
        cost_function: DataflowPlanNodeCostFunction = DefaultCostFunction[SqlDataSetT](),
        node_output_resolver: Optional[DataflowPlanNodeOutputDataSetResolver[SqlDataSetT]] = None,
        column_association_resolver: Optional[ColumnAssociationResolver] = None,
    ) -> None:
        self._data_source_semantics = semantic_model.data_source_semantics
        self._metric_semantics = semantic_model.metric_semantics
        self._metric_time_dimension_reference = DataSet.metric_time_dimension_reference()
        self._cost_function = cost_function
        self._source_nodes = source_nodes
        self._node_data_set_resolver = (
            DataflowPlanNodeOutputDataSetResolver[SqlDataSetT](
                column_association_resolver=(
                    DefaultColumnAssociationResolver(semantic_model)
                    if not column_association_resolver
                    else column_association_resolver
                ),
                semantic_model=semantic_model,
                time_spine_source=time_spine_source,
            )
            if not node_output_resolver
            else node_output_resolver
        )

    def build_plan(
        self, query_spec: MetricFlowQuerySpec, output_sql_table: Optional[SqlTable] = None
    ) -> DataflowPlan[SqlDataSetT]:
        """Generate a plan for reading the results of a query with the given spec into a dataframe or table"""
        metrics_output_node = self._build_metrics_output_node(query_spec)

        sink_node = DataflowPlanBuilder.build_sink_node_from_metrics_output_node(
            computed_metrics_output=metrics_output_node,
            order_by_specs=query_spec.order_by_specs,
            output_sql_table=output_sql_table,
            limit=query_spec.limit,
        )

        plan_id = IdGeneratorRegistry.for_class(DataflowPlanBuilder).create_id(DATAFLOW_PLAN_PREFIX)

        return DataflowPlan(plan_id=plan_id, sink_output_nodes=[sink_node])

    def _build_metrics_output_node(self, query_spec: MetricFlowQuerySpec) -> ComputedMetricsOutput[SqlDataSetT]:
        """Build a node that computes all requested metrics along with linkables."""

        compute_metrics_nodes: List[ComputedMetricsOutput[SqlDataSetT]] = []
        for metric_spec in query_spec.metric_specs:
            logger.info(f"Generating compute metrics node for {metric_spec}")
            metric = self._metric_semantics.get_metric(metric_spec)

            measure_specs = tuple(
                MeasureSpec(
                    element_name=x.element_name,
                )
                for x in metric.measure_references
            )

            logger.info(
                f"For {metric_spec}, needed measures are:\n" f"{pformat_big_objects(measure_specs=measure_specs)}"
            )

            combined_where = query_spec.where_constraint
            if metric.constraint:
                metric_constraint = MetricFlowQueryParser.convert_to_spec_where_constraint(
                    self._data_source_semantics, metric.constraint
                )
                if combined_where:
                    combined_where = combined_where.combine(metric_constraint)
                else:
                    combined_where = metric_constraint

            aggregated_measures_node = self.build_aggregated_measures(
                measure_specs=measure_specs,
                queried_linkable_specs=query_spec.linkable_specs,
                where_constraint=combined_where,
                time_range_constraint=query_spec.time_range_constraint,
                cumulative=metric.type == MetricType.CUMULATIVE,
                cumulative_window=metric.type_params.window if metric.type == MetricType.CUMULATIVE else None,
                cumulative_grain_to_date=(
                    metric.type_params.grain_to_date if metric.type == MetricType.CUMULATIVE else None
                ),
            )
            compute_metrics_nodes.append(
                self.build_computed_metrics_node(
                    metric_spec=metric_spec,
                    aggregated_measures_node=aggregated_measures_node,
                )
            )

        if len(compute_metrics_nodes) == 1:
            return compute_metrics_nodes[0]

        return CombineMetricsNode[SqlDataSetT](parent_nodes=compute_metrics_nodes)

    def build_plan_for_distinct_values(
        self,
        metric_specs: Tuple[MetricSpec, ...],
        dimension_spec: Optional[DimensionSpec] = None,
        time_dimension_spec: Optional[TimeDimensionSpec] = None,
        identifier_spec: Optional[IdentifierSpec] = None,
        time_range_constraint: Optional[TimeRangeConstraint] = None,
        limit: Optional[int] = None,
    ) -> DataflowPlan[SqlDataSetT]:
        """Generate a plan that would get the distinct values of a linkable instance.

        e.g. distinct listing__country_latest for bookings by listing__country_latest
        """
        assert_exactly_one_arg_set(
            dimension_spec=dimension_spec, time_dimension_spec=time_dimension_spec, identifier_spec=identifier_spec
        )

        # Doing this to keep the type checker happy, but assert_exactly_one_arg_set should ensure this.
        linkable_spec: Optional[LinkableInstanceSpec] = dimension_spec or time_dimension_spec or identifier_spec
        assert linkable_spec

        metrics_output_node = self._build_metrics_output_node(
            query_spec=MetricFlowQuerySpec(
                metric_specs=metric_specs,
                dimension_specs=(dimension_spec,) if dimension_spec else (),
                time_dimension_specs=(time_dimension_spec,) if time_dimension_spec else (),
                identifier_specs=(identifier_spec,) if identifier_spec else (),
                time_range_constraint=time_range_constraint,
            ),
        )

        distinct_values_node = FilterElementsNode(
            parent_node=metrics_output_node,
            include_specs=[linkable_spec],
        )

        sink_node = self.build_sink_node_from_metrics_output_node(
            computed_metrics_output=distinct_values_node,
            order_by_specs=(
                OrderBySpec(
                    item=linkable_spec,
                    descending=False,
                ),
            ),
            limit=limit,
        )

        plan_id = IdGeneratorRegistry.for_class(DataflowPlanBuilder).create_id(DATAFLOW_PLAN_PREFIX)

        return DataflowPlan(
            plan_id=plan_id,
            sink_output_nodes=[sink_node],
        )

    @staticmethod
    def build_sink_node_from_metrics_output_node(
        computed_metrics_output: BaseOutput[SqlDataSetT],
        order_by_specs: Sequence[OrderBySpec],
        output_sql_table: Optional[SqlTable] = None,
        limit: Optional[int] = None,
    ) -> SinkOutput[SqlDataSetT]:
        """Adds order by / limit / write nodes."""
        order_by_limit: Optional[OrderByLimitNode[SqlDataSetT]] = None

        if order_by_specs or limit:
            order_by_limit = OrderByLimitNode[SqlDataSetT](
                order_by_specs=list(order_by_specs),
                limit=limit,
                parent_node=computed_metrics_output,
            )

        write_result_node: SinkOutput[SqlDataSetT]
        if not output_sql_table:
            write_result_node = WriteToResultDataframeNode[SqlDataSetT](
                parent_node=order_by_limit or computed_metrics_output,
            )
        else:
            write_result_node = WriteToResultTableNode[SqlDataSetT](
                parent_node=order_by_limit or computed_metrics_output,
                output_sql_table=output_sql_table,
            )

        return write_result_node

    @staticmethod
    def _contains_multihop_linkables(linkable_specs: Sequence[LinkableInstanceSpec]) -> bool:
        """Returns true if any of the linkable specs requires a multi-hop join to realize."""
        return any(len(x.identifier_links) > 1 for x in linkable_specs)

    def _get_data_source_names_for_measures(self, measure_names: Sequence[MeasureSpec]) -> Set[str]:
        """Return the names of the data sources needed to compute the input measures

        This is a temporary method for use in assertion boundaries while we implement support for multiple data sources
        """
        data_source_names: Set[str] = set()
        for measure_name in measure_names:
            data_source_names = data_source_names.union(
                {d.name for d in self._data_source_semantics.get_data_sources_for_measure(measure_name.as_reference)}
            )
        return data_source_names

    def _sort_by_suitability(self, nodes: Sequence[BaseOutput[SqlDataSetT]]) -> Sequence[BaseOutput[SqlDataSetT]]:
        """Sort nodes by the cost, then by the number of linkable specs.

        Lower cost nodes will result in faster queries, and the lower the number of linkable specs means less
        aggregation required.
        """

        def sort_function(node: BaseOutput[SqlDataSetT]) -> Tuple[int, int]:
            data_set = self._node_data_set_resolver.get_output_data_set(node)
            return self._cost_function.calculate_cost(node).as_int, len(data_set.instance_set.spec_set.linkable_specs)

        return sorted(nodes, key=sort_function)

    def _select_source_nodes_with_meausres(
        self, measure_specs: Set[MeasureSpec], source_nodes: Sequence[BaseOutput[SqlDataSetT]]
    ) -> Sequence[BaseOutput[SqlDataSetT]]:
        nodes = []
        measure_specs_set = set(measure_specs)
        for source_node in source_nodes:
            measure_specs_in_node = self._node_data_set_resolver.get_output_data_set(
                source_node
            ).instance_set.spec_set.measure_specs
            if measure_specs_set.intersection(set(measure_specs_in_node)) == measure_specs_set:
                nodes.append(source_node)
        return nodes

    def _find_measure_recipe(
        self,
        measure_specs: Sequence[MeasureSpec],
        linkable_specs: Sequence[LinkableInstanceSpec],
        time_range_constraint: Optional[TimeRangeConstraint] = None,
    ) -> Optional[MeasureRecipe]:
        """Find a recipe for getting measure_specs along with the linkable specs.

        Prior to calling this method we should always be checking that all input measure specs come from
        the same base data source, otherwise the internal conditions here will be impossible to satisfy
        """
        data_sources = self._get_data_source_names_for_measures(measure_specs)
        if len(data_sources) > 1:
            raise ValueError(
                f"Cannot find a single measure recipe for measures {measure_specs} coming from multiple "
                f"data sources: {data_sources}. This suggests the measure_specs were not correctly filtered."
            )

        node_processor = PreDimensionJoinNodeProcessor(
            data_source_semantics=self._data_source_semantics,
            node_data_set_resolver=self._node_data_set_resolver,
        )

        source_nodes: Sequence[BaseOutput[SqlDataSetT]] = self._source_nodes

        # We only care about nodes that have all required measures
        potential_measure_nodes: Sequence[BaseOutput[SqlDataSetT]] = self._select_source_nodes_with_meausres(
            measure_specs=set(measure_specs), source_nodes=source_nodes
        )

        logger.info(f"There are {len(potential_measure_nodes)} potential measure source nodes")

        logger.info(f"Starting search with {len(source_nodes)} source nodes")
        start_time = time.time()
        # Only apply the time constraint to nodes that will be used for measures because some dimensional sources have
        # measures in them, and time constraining those would result in incomplete joins.
        if time_range_constraint:
            potential_measure_nodes = node_processor.add_time_range_constraint(
                source_nodes=potential_measure_nodes,
                metric_time_dimension_reference=self._metric_time_dimension_reference,
                time_range_constraint=time_range_constraint,
            )

        nodes_available_for_joins = node_processor.remove_unnecessary_nodes(
            desired_linkable_specs=linkable_specs,
            nodes=source_nodes,
            metric_time_dimension_reference=self._metric_time_dimension_reference,
        )
        logger.info(
            f"After removing unnecessary nodes, there are {len(nodes_available_for_joins)} nodes available for joins"
        )
        if DataflowPlanBuilder._contains_multihop_linkables(linkable_specs):
            nodes_available_for_joins = node_processor.add_multi_hop_joins(linkable_specs, source_nodes)
            logger.info(
                f"After adding multi-hop nonds, there are {nodes_available_for_joins} nodes available for joins"
            )

        logger.info(f"Processing nodes took: {time.time()-start_time:.2f}s")

        node_evaluator = NodeEvaluatorForLinkableInstances(
            data_source_semantics=self._data_source_semantics,
            nodes_available_for_joins=self._sort_by_suitability(nodes_available_for_joins),
            node_data_set_resolver=self._node_data_set_resolver,
        )

        # Dict from the node that contains the measure spec to the evaluation results.
        node_to_evaluation: Dict[BaseOutput, LinkableInstanceSatisfiabilityEvaluation] = {}

        for node in self._sort_by_suitability(potential_measure_nodes):
            data_set = self._node_data_set_resolver.get_output_data_set(node)

            missing_specs = [spec for spec in measure_specs if spec not in data_set.instance_set.spec_set.measure_specs]
            if missing_specs:
                logger.debug(
                    f"Skipping evaluation for node since it does not have all of the measure specs {missing_specs}:"
                    f"\n\n{dataflow_dag_as_text(node)}"
                )
                continue

            logger.debug(f"Evaluating measure node:\n{pformat_big_objects(measure_node=dataflow_dag_as_text(node))}")

            start_time = time.time()
            evaluation = node_evaluator.evaluate_node(
                start_node=node,
                required_linkable_specs=list(linkable_specs),
            )
            logger.info(f"Evaluation of {node} took {time.time() - start_time}s")

            logger.debug(
                f"Evaluation for measure node is:\n"
                f"{pformat_big_objects(node=dataflow_dag_as_text(node), evaluation=evaluation)}"
            )

            if len(evaluation.unjoinable_linkable_specs) > 0:
                logger.debug(
                    f"Skipping {node.node_id} since it contains un-joinable specs: "
                    f"{evaluation.unjoinable_linkable_specs}"
                )
                continue

            num_joins_required = len(evaluation.join_recipes)
            logger.info(f"Found candidate with node ID '{node.node_id}' with {num_joins_required} joins required.")

            node_to_evaluation[node] = evaluation

            # Since are evaluating nodes with the lowest cost first, if we find one without requiring any joins, then
            # this is going to be the lowest cost solution.
            if len(evaluation.join_recipes) == 0:
                logger.info("Not evaluating other nodes since we found one that doesn't require joins")

        logger.info(f"Found {len(node_to_evaluation)} candidate measure nodes.")

        if len(node_to_evaluation) > 0:

            cost_function = DefaultCostFunction[SqlDataSetT]()

            node_with_lowest_cost = min(node_to_evaluation, key=cost_function.calculate_cost)
            evaluation = node_to_evaluation[node_with_lowest_cost]
            logger.info(
                "Lowest cost node is:\n"
                + pformat_big_objects(
                    lowest_cost_node=dataflow_dag_as_text(node_with_lowest_cost),
                    evaluation=evaluation,
                    cost=cost_function.calculate_cost(node_with_lowest_cost),
                )
            )

            # Nodes containing the linkable instances will be joined to the node containing the measure, so these
            # identifiers will need to be present in the measure node.
            required_local_identifier_specs = tuple(x.join_on_identifier for x in evaluation.join_recipes)
            # Same thing with partitions.
            required_local_dimension_specs = tuple(
                y.start_node_dimension_spec for x in evaluation.join_recipes for y in x.join_on_partition_dimensions
            )
            required_local_time_dimension_specs = tuple(
                y.start_node_time_dimension_spec
                for x in evaluation.join_recipes
                for y in x.join_on_partition_time_dimensions
            )

            return MeasureRecipe(
                measure_node=node_with_lowest_cost,
                required_local_linkable_specs=(
                    evaluation.local_linkable_specs
                    + required_local_identifier_specs
                    + required_local_dimension_specs
                    + required_local_time_dimension_specs
                ),
                join_linkable_instances_recipes=node_to_evaluation[node_with_lowest_cost].join_recipes,
            )

        logger.error("No recipe could be constructed.")
        return None

    def build_computed_metrics_node(
        self,
        metric_spec: MetricSpec,
        aggregated_measures_node: Union[AggregateMeasuresNode[SqlDataSetT], BaseOutput[SqlDataSetT]],
    ) -> ComputeMetricsNode[SqlDataSetT]:
        """Builds a ComputeMetricsNode from aggregated measures."""

        return ComputeMetricsNode[SqlDataSetT](
            parent_node=aggregated_measures_node,
            metric_specs=[metric_spec],
        )

    def build_aggregated_measures(
        self,
        measure_specs: Tuple[MeasureSpec, ...],
        queried_linkable_specs: LinkableSpecSet,
        where_constraint: Optional[SpecWhereClauseConstraint] = None,
        time_range_constraint: Optional[TimeRangeConstraint] = None,
        cumulative: Optional[bool] = False,
        cumulative_window: Optional[CumulativeMetricWindow] = None,
        cumulative_grain_to_date: Optional[TimeGranularity] = None,
    ) -> BaseOutput[SqlDataSetT]:
        """Returns a node where the measures are aggregated by the linkable specs and constrained appropriately.

        This might be a node representing a single aggregation over one data source, or a node representing
        a composite set of aggregations originating from multiple data sources, and joined into a single
        aggregated set of measures.
        """
        output_nodes: List[BaseOutput[SqlDataSetT]] = []
        data_sources_for_measures = collections.defaultdict(list)
        for measure_spec in measure_specs:
            data_source_names = [
                dsource.name
                for dsource in self._data_source_semantics.get_data_sources_for_measure(
                    measure_reference=measure_spec.as_reference
                )
            ]
            assert (
                len(data_source_names) == 1
            ), f"Validation should enforce one data source per measure, but found {data_source_names} for {measure_spec}!"
            data_sources_for_measures[data_source_names[0]].append(measure_spec)

        for data_source, measures in data_sources_for_measures.items():
            logger.info(f"Building aggregated measures for {data_source}. Measures: {measures}")
            node = self._build_aggregated_measures_for_single_data_source(
                measure_specs=tuple(measures),
                queried_linkable_specs=queried_linkable_specs,
                where_constraint=where_constraint,
                time_range_constraint=time_range_constraint,
                cumulative=cumulative,
                cumulative_window=cumulative_window,
                cumulative_grain_to_date=cumulative_grain_to_date,
            )
            output_nodes.append(node)

        if len(output_nodes) == 1:
            return output_nodes[0]
        else:
            return FilterElementsNode(
                parent_node=JoinAggregatedMeasuresByGroupByColumnsNode(parent_nodes=output_nodes),
                include_specs=LinkableInstanceSpec.merge(queried_linkable_specs.as_tuple, measure_specs),
            )

    @staticmethod
    def _add_where_and_aggregate(
        measure_specs: Tuple[MeasureSpec, ...],
        queried_linkable_specs: LinkableSpecSet,
        unaggregated_node: BaseOutput[SqlDataSetT],
        where_constraint: SpecWhereClauseConstraint,
    ) -> BaseOutput[SqlDataSetT]:
        """Applies the where constraint appropriately and produce aggregated measures.

        Assumes that the input node has all measure specs, queried_linkable_specs, and also the specs required by the
        where clause.
        """
        where_constrained_node = WhereConstraintNode(
            parent_node=unaggregated_node,
            where_constraint=where_constraint,
        )

        if where_constraint.linkable_spec_set.is_subset_of(queried_linkable_specs):
            return AggregateMeasuresNode[SqlDataSetT](where_constrained_node)

        # At this point, it's the case that the linkable specs in the where clause are not a subset of the queried
        # linkable specs. A filter is needed after a where clause so that the linkable specs in the where clause don't
        # show up in the final result.
        #
        # e.g. for "bookings" by "ds" where "is_instant", "is_instant" should not be in the results.
        filtered_node = FilterElementsNode(
            parent_node=where_constrained_node,
            include_specs=measure_specs + queried_linkable_specs.as_tuple,
        )

        return AggregateMeasuresNode[SqlDataSetT](parent_node=filtered_node)

    def _build_aggregated_measures_for_single_data_source(
        self,
        measure_specs: Tuple[MeasureSpec, ...],
        queried_linkable_specs: LinkableSpecSet,
        where_constraint: Optional[SpecWhereClauseConstraint] = None,
        time_range_constraint: Optional[TimeRangeConstraint] = None,
        cumulative: Optional[bool] = False,
        cumulative_window: Optional[CumulativeMetricWindow] = None,
        cumulative_grain_to_date: Optional[TimeGranularity] = None,
    ) -> BaseOutput[SqlDataSetT]:
        metric_time_dimension_requested = self._metric_time_dimension_reference.element_name in [
            linkable_spec.element_name for linkable_spec in queried_linkable_specs.as_tuple
        ]
        cumulative_metric_adjusted_time_constraint: Optional[TimeRangeConstraint] = None
        if cumulative and time_range_constraint is not None:
            logger.info(f"Time range constraint before adjustment is {time_range_constraint}")
            granularity: Optional[TimeGranularity] = None
            count = 0
            if cumulative_window is not None:
                granularity = cumulative_window.granularity
                count = cumulative_window.count
            elif cumulative_grain_to_date is not None:
                count = 1
                granularity = cumulative_grain_to_date

            cumulative_metric_adjusted_time_constraint = (
                time_range_constraint.adjust_time_constraint_for_cumulative_metric(granularity, count)
            )
            logger.info(f"Adjusted time range constraint {cumulative_metric_adjusted_time_constraint}")

        required_linkable_specs = (
            LinkableSpecSet.merge((queried_linkable_specs, where_constraint.linkable_spec_set))
            if where_constraint
            else queried_linkable_specs
        )

        logger.info(
            f"Looking for a recipe to get:\n"
            f"{pformat_big_objects(measure_specs=measure_specs, required_linkable_set=required_linkable_specs)}"
        )

        find_recipe_start_time = time.time()
        measure_recipe = self._find_measure_recipe(
            measure_specs=measure_specs,
            time_range_constraint=cumulative_metric_adjusted_time_constraint or time_range_constraint,
            linkable_specs=required_linkable_specs.as_tuple,
        )
        logger.info(
            f"With {len(self._source_nodes)} source nodes, finding a recipe took "
            f"{time.time() - find_recipe_start_time:.2f}s"
        )

        logger.info(f"Using recipe:\n{pformat_big_objects(measure_recipe=measure_recipe)}")

        if not measure_recipe:
            # TODO: Improve for better user understandability.
            raise UnableToSatisfyQueryError(
                f"Recipe not found for measure specs: {measure_specs} and linkable specs: {required_linkable_specs}"
            )

        # Only get the required measure and the local linkable instances so that aggregations work correctly.
        filtered_measure_source_node = FilterElementsNode[SqlDataSetT](
            parent_node=measure_recipe.measure_node,
            include_specs=InstanceSpec.merge(measure_specs, measure_recipe.required_local_linkable_specs),
        )

        time_range_node: Optional[JoinOverTimeRangeNode[SqlDataSetT]] = None
        if cumulative:
            time_range_node = JoinOverTimeRangeNode(
                parent_node=filtered_measure_source_node,
                metric_time_dimension_reference=self._metric_time_dimension_reference,
                window=cumulative_window,
                grain_to_date=cumulative_grain_to_date,
                time_range_constraint=time_range_constraint,
            )

        filtered_measure_or_time_range_node = time_range_node or filtered_measure_source_node
        join_targets = []

        for join_recipe in measure_recipe.join_linkable_instances_recipes:
            # Figure out what elements to filter from the joined node.

            # Sanity check - all linkable specs should have a link, or else why would we be joining them.
            assert all([len(x.identifier_links) > 0 for x in join_recipe.satisfiable_linkable_specs])

            # If we're joining something in, then we need the associated identifier and partitions.
            include_specs: List[LinkableInstanceSpec] = [
                x.identifier_links[0] for x in join_recipe.satisfiable_linkable_specs
            ]
            include_specs.extend([x.node_to_join_dimension_spec for x in join_recipe.join_on_partition_dimensions])
            include_specs.extend(
                [x.node_to_join_time_dimension_spec for x in join_recipe.join_on_partition_time_dimensions]
            )

            # satisfiable_linkable_specs describes what can be satisfied after the join, so remove the identifier
            # link when filtering before the join.
            # e.g. if the node is used to satisfy "user_id__country", then the node must have the identifier
            # "user_id" and the "country" dimension so that it can be joined to the measure node.
            include_specs.extend([x.without_first_identifier_link() for x in join_recipe.satisfiable_linkable_specs])
            filtered_node_to_join = FilterElementsNode[SqlDataSetT](
                parent_node=join_recipe.node_to_join,
                include_specs=include_specs,
            )
            join_targets.append(
                JoinDescription(
                    join_node=filtered_node_to_join,
                    join_on_identifier=join_recipe.join_on_identifier,
                    join_on_partition_dimensions=join_recipe.join_on_partition_dimensions,
                    join_on_partition_time_dimensions=join_recipe.join_on_partition_time_dimensions,
                )
            )

        unaggregated_measure_node: BaseOutput[SqlDataSetT]
        if len(join_targets) > 0:
            filtered_measures_with_joined_elements = JoinToBaseOutputNode[SqlDataSetT](
                parent_node=filtered_measure_or_time_range_node,
                join_targets=join_targets,
            )

            specs_to_keep_after_join: List[InstanceSpec] = list(measure_specs)
            specs_to_keep_after_join.extend(required_linkable_specs.as_tuple)

            after_join_filtered_node = FilterElementsNode[SqlDataSetT](
                parent_node=filtered_measures_with_joined_elements,
                include_specs=specs_to_keep_after_join,
            )
            unaggregated_measure_node = after_join_filtered_node
        else:
            unaggregated_measure_node = filtered_measure_or_time_range_node

        cumulative_metric_constrained_node: Optional[ConstrainTimeRangeNode] = None
        if (
            cumulative_metric_adjusted_time_constraint is not None
            and time_range_constraint is not None
            and metric_time_dimension_requested
        ):
            cumulative_metric_constrained_node = ConstrainTimeRangeNode(
                unaggregated_measure_node, time_range_constraint
            )

        if where_constraint:
            return DataflowPlanBuilder._add_where_and_aggregate(
                measure_specs=measure_specs,
                unaggregated_node=cumulative_metric_constrained_node or unaggregated_measure_node,
                queried_linkable_specs=queried_linkable_specs,
                where_constraint=where_constraint,
            )

        return AggregateMeasuresNode[SqlDataSetT](
            parent_node=cumulative_metric_constrained_node or unaggregated_measure_node
        )



"""Functions to help generate a text representation of a dataflow plan."""

import logging
import textwrap

import jinja2

from metricflow.dag.dag_to_text import MetricFlowDagToText
from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
    DataflowPlan,
)

logger = logging.getLogger(__name__)


def dataflow_dag_as_text(root_node: DataflowPlanNode) -> str:
    """Converts the dataflow dag starting from the given root node to a text representation.

    The text representation is similar to XML.
    """

    return MetricFlowDagToText().to_text(root_node)


def dataflow_plan_as_text(dataflow_plan: DataflowPlan) -> str:
    """Converts the dataflow plan to a text representation that can be used for tests / debugging.

    The text representation is similar to XML.
    """

    # Convert each of the components that are associated with the sink nodes to a text representation.
    component_from_sink_nodes_as_text = []
    for sink_node in dataflow_plan.sink_nodes:
        component_from_sink_nodes_as_text.append(dataflow_dag_as_text(sink_node))

    # Under <DataflowPlan>, render all components.
    return jinja2.Template(
        textwrap.dedent(
            """\
            <{{ node_class }}{%- if not inner_contents %}/>{%- else %}>
                {%- if inner_contents %}
                {{ inner_contents | indent(4) }}
                {%- endif %}
            </{{ node_class }}>
            {%- endif %}
            """
        ),
        undefined=jinja2.StrictUndefined,
    ).render(
        node_class=dataflow_plan.__class__.__name__,
        inner_contents="\n".join(component_from_sink_nodes_as_text),
    )



from datetime import date
from typing import List

import pytest
from metricflow.instances import (
    DataSourceElementReference,
    DataSourceReference,
    MaterializationModelReference,
    MetricModelReference,
)

from metricflow.model.validations.validator_helpers import (
    DataSourceContext,
    DataSourceElementContext,
    DataSourceElementType,
    FileContext,
    MaterializationContext,
    MetricContext,
    ModelValidationResults,
    ValidationError,
    ValidationFutureError,
    ValidationIssueLevel,
    ValidationIssueType,
    ValidationWarning,
)


@pytest.fixture
def list_of_issues() -> List[ValidationIssueType]:  # noqa: D
    file_context = FileContext(file_name="foo", line_number=1337)
    data_source_name = "My data source"

    issues: List[ValidationIssueType] = []
    issues.append(
        ValidationWarning(
            context=DataSourceContext(
                file_context=file_context,
                data_source=DataSourceReference(data_source_name=data_source_name),
            ),
            message="Something caused a warning, problem #1",
        )
    )
    issues.append(
        ValidationWarning(
            context=DataSourceElementContext(
                file_context=file_context,
                data_source_element=DataSourceElementReference(
                    data_source_name=data_source_name, element_name="My dimension"
                ),
                element_type=DataSourceElementType.DIMENSION,
            ),
            message="Something caused a warning, problem #2",
        )
    )
    issues.append(
        ValidationFutureError(
            context=DataSourceElementContext(
                file_context=file_context,
                data_source_element=DataSourceElementReference(
                    data_source_name=data_source_name, element_name="My identifier"
                ),
                element_type=DataSourceElementType.IDENTIFIER,
            ),
            message="Something caused a future error, problem #3",
            error_date=date(2022, 6, 13),
        )
    )
    issues.append(
        ValidationError(
            context=DataSourceElementContext(
                file_context=file_context,
                data_source_name=data_source_name,
                data_source_element=DataSourceElementReference(
                    data_source_name=data_source_name, element_name="My measure"
                ),
                element_type=DataSourceElementType.MEASURE,
            ),
            message="Something caused an error, problem #4",
        )
    )
    issues.append(
        ValidationError(
            context=MaterializationContext(
                file_context=file_context,
                materialization=MaterializationModelReference(materialization_name="My materialization"),
            ),
            message="Something caused an error, problem #5",
        )
    )
    issues.append(
        ValidationError(
            context=MetricContext(
                file_context=file_context,
                metric=MetricModelReference(metric_name="My metric"),
            ),
            message="Something caused a error, problem #6",
        )
    )
    issues.append(ValidationError(context=file_context, message="Something caused a error, probelm #7"))
    return issues


def test_creating_model_validation_results_from_issue_list(  # noqa: D
    list_of_issues: List[ValidationIssueType],
) -> None:
    warnings = [issue for issue in list_of_issues if issue.level == ValidationIssueLevel.WARNING]
    future_errors = [issue for issue in list_of_issues if issue.level == ValidationIssueLevel.FUTURE_ERROR]
    errors = [issue for issue in list_of_issues if issue.level == ValidationIssueLevel.ERROR]

    model_validation_issues = ModelValidationResults.from_issues_sequence(list_of_issues)
    assert len(model_validation_issues.warnings) == len(warnings)
    assert len(model_validation_issues.future_errors) == len(future_errors)
    assert len(model_validation_issues.errors) == len(errors)
    assert model_validation_issues.has_blocking_issues

    model_validation_issues = ModelValidationResults(warnings=warnings, future_errors=future_errors)
    assert not model_validation_issues.has_blocking_issues


def test_jsonifying_and_reloading_model_validation_results_is_equal(  # noqa: D
    list_of_issues: List[ValidationIssueType],
) -> None:
    warnings = [issue for issue in list_of_issues if issue.level == ValidationIssueLevel.WARNING]
    errors = [issue for issue in list_of_issues if issue.level == ValidationIssueLevel.ERROR]
    set_context_types = set([issue.context.__class__ for issue in list_of_issues])

    model_validation_issues = ModelValidationResults.from_issues_sequence(list_of_issues)
    model_validation_issues_new = ModelValidationResults.parse_raw(model_validation_issues.json())
    assert model_validation_issues_new == model_validation_issues
    assert model_validation_issues_new != ModelValidationResults(warnings=warnings, errors=errors)

    # ensure ValidationContexts were properly parsed into the differen subclasses
    new_context_types = [issue.context.__class__ for issue in model_validation_issues_new.warnings]
    new_context_types += [issue.context.__class__ for issue in model_validation_issues_new.future_errors]
    new_context_types += [issue.context.__class__ for issue in model_validation_issues_new.errors]
    assert set_context_types == set(new_context_types)


def test_merge_two_model_validation_results(list_of_issues: List[ValidationIssueType]) -> None:  # noqa: D
    validation_results = ModelValidationResults.from_issues_sequence(list_of_issues)
    validation_results_dup = ModelValidationResults.from_issues_sequence(list_of_issues)
    merged = ModelValidationResults.merge([validation_results, validation_results_dup])

    assert merged.warnings == validation_results.warnings + validation_results_dup.warnings
    assert merged.future_errors == validation_results.future_errors + validation_results_dup.future_errors
    assert merged.errors == validation_results.errors + validation_results_dup.errors



import datetime as dt

from metricflow.time.time_source import TimeSource


class ServerTimeSource(TimeSource):
    """A time source that represents the current datetime in UTC."""

    def get_time(self) -> dt.datetime:  # noqa: D
        return dt.datetime.utcnow()



import logging
from typing import Generic, Tuple, Optional, Union

from metricflow.protocols.sql_client import SqlClient
from metricflow.dag.id_generation import IdGeneratorRegistry, SQL_QUERY_PLAN_PREFIX, EXEC_PLAN_PREFIX
from metricflow.dataflow.dataflow_plan import (
    SourceDataSetT,
    WriteToResultDataframeNode,
    DataflowPlan,
    SinkNodeVisitor,
    WriteToResultTableNode,
    ComputedMetricsOutput,
    BaseOutput,
)
from metricflow.dataflow.sql_table import SqlTable
from metricflow.execution.execution_plan import (
    ExecutionPlan,
    SelectSqlQueryToDataFrameTask,
    ExecutionPlanTask,
    SelectSqlQueryToTableTask,
)
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter, SqlDataSetT
from metricflow.specs import OutputColumnNameOverride
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.sql.sql_plan import SqlSelectStatementNode, SqlSelectColumn, SqlQueryPlan
from metricflow.sql.sql_plan_to_text import sql_query_plan_as_text

logger = logging.getLogger(__name__)


class DataflowToExecutionPlanConverter(Generic[SqlDataSetT], SinkNodeVisitor[SqlDataSetT, ExecutionPlan]):
    """Converts a dataflow plan to an execution plan"""

    def __init__(
        self,
        sql_plan_converter: DataflowToSqlQueryPlanConverter[SqlDataSetT],
        sql_plan_renderer: SqlQueryPlanRenderer,
        sql_client: SqlClient,
        output_column_name_overrides: Tuple[OutputColumnNameOverride, ...] = (),
    ) -> None:
        """Constructor.

        Args:
            sql_plan_converter: Converts a dataflow plan node to a SQL query plan
            sql_plan_renderer: Converts a SQL query plan to SQL text
            sql_client: The client to use for running queries.
            output_column_name_overrides: In the output dataframe / table, name output columns in a specific way.
        """
        self._sql_plan_converter = sql_plan_converter
        self._sql_plan_renderer = sql_plan_renderer
        self._sql_client = sql_client
        self._output_column_name_overrides = output_column_name_overrides

    @staticmethod
    def override_output_column_names(
        sql_plan_converter: DataflowToSqlQueryPlanConverter[SqlDataSetT],
        output_column_name_overrides: Tuple[OutputColumnNameOverride, ...],
        select_node: SqlSelectStatementNode,
    ) -> SqlSelectStatementNode:
        """Change the output column names in the select_node according ot output_column_name_overrides."""
        column_name_mapping = {}
        for column_name_override in output_column_name_overrides:
            expected_column_name = sql_plan_converter.column_association_resolver.resolve_time_dimension_spec(
                column_name_override.time_dimension_spec
            ).column_name
            assert expected_column_name in [
                x.column_alias for x in select_node.select_columns
            ], f"Column {expected_column_name} not in {[x.column_alias for x in select_node.select_columns]}"
            new_column_name = column_name_override.output_column_name
            column_name_mapping[expected_column_name] = new_column_name

        new_select_columns = []
        for select_column in select_node.select_columns:
            if select_column.column_alias in column_name_mapping:
                new_select_columns.append(
                    SqlSelectColumn(
                        expr=select_column.expr, column_alias=column_name_mapping[select_column.column_alias]
                    )
                )
            else:
                new_select_columns.append(select_column)

        return SqlSelectStatementNode(
            description=select_node.description,
            select_columns=tuple(new_select_columns),
            from_source=select_node.from_source,
            from_source_alias=select_node.from_source_alias,
            joins_descs=select_node.join_descs,
            group_bys=select_node.group_bys,
            order_bys=select_node.order_bys,
            where=select_node.where,
            limit=select_node.limit,
        )

    def _build_execution_plan(  # noqa: D
        self,
        node: Union[BaseOutput[SourceDataSetT], ComputedMetricsOutput[SourceDataSetT]],
        output_table: Optional[SqlTable] = None,
    ) -> ExecutionPlan:

        sql_plan = self._sql_plan_converter.convert_to_sql_query_plan(
            sql_engine_attributes=self._sql_client.sql_engine_attributes,
            sql_query_plan_id=IdGeneratorRegistry.for_class(SqlQueryPlan).create_id(SQL_QUERY_PLAN_PREFIX),
            dataflow_plan_node=node,
        )

        if self._output_column_name_overrides:
            select_node = sql_plan.render_node.as_select_node
            assert select_node

            sql_plan = SqlQueryPlan(
                plan_id=IdGeneratorRegistry.for_class(SqlQueryPlan).create_id(SQL_QUERY_PLAN_PREFIX),
                render_node=DataflowToExecutionPlanConverter.override_output_column_names(
                    sql_plan_converter=self._sql_plan_converter,
                    output_column_name_overrides=self._output_column_name_overrides,
                    select_node=select_node,
                ),
            )

        logger.debug(f"Generated SQL query plan is:\n{sql_query_plan_as_text(sql_plan)}")

        render_result = self._sql_plan_renderer.render_sql_query_plan(sql_plan)

        leaf_task: ExecutionPlanTask

        if not output_table:
            leaf_task = SelectSqlQueryToDataFrameTask(
                sql_client=self._sql_client,
                sql_query=render_result.sql,
                execution_parameters=render_result.execution_parameters,
            )
        else:
            leaf_task = SelectSqlQueryToTableTask(
                sql_client=self._sql_client,
                sql_query=render_result.sql,
                execution_parameters=render_result.execution_parameters,
                output_table=output_table,
            )

        return ExecutionPlan(
            plan_id=IdGeneratorRegistry.for_class(self.__class__).create_id(EXEC_PLAN_PREFIX), leaf_tasks=[leaf_task]
        )

    def visit_write_to_result_dataframe_node(  # noqa: D
        self, node: WriteToResultDataframeNode[SourceDataSetT]
    ) -> ExecutionPlan:
        logger.info(f"Generating SQL query plan from {node.node_id} -> {node.parent_node.node_id}")
        return self._build_execution_plan(node.parent_node)

    def visit_write_to_result_table_node(  # noqa: D
        self, node: WriteToResultTableNode[SourceDataSetT]
    ) -> ExecutionPlan:
        logger.info(f"Generating SQL query plan from {node.node_id} -> {node.parent_node.node_id}")
        return self._build_execution_plan(node.parent_node, node.output_sql_table)

    def convert_to_execution_plan(self, dataflow_plan: DataflowPlan) -> ExecutionPlan:
        """Convert the dataflow plan to an execution plan."""

        assert len(dataflow_plan.sink_output_nodes) == 1, "Only 1 sink node in the plan is currently supported."
        return dataflow_plan.sink_output_nodes[0].accept_sink_node_visitor(self)



from typing import Dict, TypeVar, Generic

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter
from metricflow.plan_conversion.sql_dataset import SqlDataSet
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.specs import ColumnAssociationResolver
from metricflow.model.semantic_model import SemanticModel

SourceDataSetT = TypeVar("SourceDataSetT", bound=SqlDataSet)


class DataflowPlanNodeOutputDataSetResolver(Generic[SourceDataSetT], DataflowToSqlQueryPlanConverter[SourceDataSetT]):
    """Given a node in a dataflow plan, figure out what is the data set output by that node.

    Recall that in the dataflow plan, the nodes represent computation, and the inputs and outputs of the nodes are
    data sets containing measures, dimensions, etc.

    Note: the term "dimension" is used below, but it actually refers to any LinkableInstance.

    This class is useful when computing the joins required to retrieve the dimensions associated with a measure. The
    object that figures out the joins (NodeEvaluatorForLinkableInstances) is given a set of dataflow plan nodes that
    output different dimensions (this information is not necessarily in the node itself).

    NodeEvaluatorForLinkableInstances needs the data set associated with a node so that it knows what dimensions and
    identifiers are in the node. This information is used to figure out whether that node is useful to join to in order
    to retrieve a particular dimension.

    In the simple case, if the input nodes are all ReadSqlSourceNodes, then the data set is a member variable of the
    node, and there would be no need to traverse the dataflow node hierarchy with this class. However, if it's not a
    single node, that information is not available as the dataflow plan nodes contain little metadata. In that case,
    this class is needed to traverse the dataflow plan and figure that out.

    The reason why the input nodes to NodeEvaluatorForLinkableInstances wouldn't always be all ReadSqlSourceNodes is to
    realize an easy way to handle multi-hop joins without making changes to NodeEvaluatorForLinkableInstances. If we
    generate a set of nodes that already include the multi-hop dimensions available, the join resolution logic becomes
    much simpler. For example, a node like:

    <JoinToBaseOutputNode>
        <!-- Join dim_users and dim_devices by device_id -->
        <ReadSqlSourceNode>
          <!-- Read from dim_users to get user_id, device_id -->
        <ReadSqlSourceNodes>
          <!-- Read from dim_devices device_id, platform -->
    </JoinToBaseOutputNode>

    would have the dimension user_id__device_id__platform available, so to NodeEvaluatorForLinkableInstances,
    it's the same problem as doing a single-hop join. This simplifies the join resolution logic, though now the input
    to NodeEvaluatorForLinkableInstances needs to contain nodes that include all possible multi-hop joins.

    The logic to figure the dataset output by a node is the same as DataflowToSqlQueryPlanConverter because the same
    information is needed for generating SQL queries, so inheriting from that. We may want to look later at making
    another class to have better separation of concerns.
    """

    def __init__(  # noqa: D
        self,
        column_association_resolver: ColumnAssociationResolver,
        semantic_model: SemanticModel,
        time_spine_source: TimeSpineSource,
    ) -> None:
        self._node_to_output_data_set: Dict[DataflowPlanNode[SourceDataSetT], SqlDataSet] = {}
        super().__init__(
            column_association_resolver=column_association_resolver,
            semantic_model=semantic_model,
            time_spine_source=time_spine_source,
        )

    def get_output_data_set(self, node: DataflowPlanNode[SourceDataSetT]) -> SqlDataSet:  # noqa: D
        """Cached since this will be called repeatedly during the computation of multiple metrics."""

        if node not in self._node_to_output_data_set:
            self._node_to_output_data_set[node] = node.accept(self)

        return self._node_to_output_data_set[node]



import logging

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.transformations.transform_rule import ModelTransformRule

logger = logging.getLogger(__name__)


class CompositeIdentifierExpressionRule(ModelTransformRule):
    """Transform composite sub-identifiers for convenience.

    If a sub-identifier has no expression, check if an identifier exists with the same name and use that identifier's
    expression if it has one.
    """

    @staticmethod
    def transform_model(model: UserConfiguredModel) -> UserConfiguredModel:  # noqa: D
        for data_source in model.data_sources:
            for identifier in data_source.identifiers:
                if identifier.identifiers is None or len(identifier.identifiers) == 0:
                    continue

                for sub_identifier in identifier.identifiers:
                    if sub_identifier.name or sub_identifier.expr:
                        continue

                    for identifier in data_source.identifiers:
                        if sub_identifier.ref == identifier.name:
                            sub_identifier.ref = None
                            sub_identifier.name = identifier.name
                            sub_identifier.expr = identifier.expr
                            break

        return model



import logging

from metricflow.dataflow.builder.costing import DefaultCostFunction, DefaultCost
from metricflow.dataflow.dataflow_plan import (
    FilterElementsNode,
    AggregateMeasuresNode,
    JoinToBaseOutputNode,
    JoinDescription,
)
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.specs import (
    MeasureSpec,
    IdentifierSpec,
    DimensionSpec,
    LinklessIdentifierSpec,
)
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository

logger = logging.getLogger(__name__)


def test_costing(consistent_id_object_repository: ConsistentIdObjectRepository) -> None:  # noqa: D
    bookings_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    listings_node = consistent_id_object_repository.simple_model_read_nodes["listings_latest"]

    bookings_filtered = FilterElementsNode[DataSourceDataSet](
        parent_node=bookings_node,
        include_specs=[
            MeasureSpec(
                element_name="bookings",
            ),
            IdentifierSpec(
                element_name="listing",
                identifier_links=(),
            ),
        ],
    )

    listings_filtered = FilterElementsNode[DataSourceDataSet](
        parent_node=listings_node,
        include_specs=[
            DimensionSpec(
                element_name="country_latest",
                identifier_links=(),
            ),
            IdentifierSpec(
                element_name="listing",
                identifier_links=(),
            ),
        ],
    )

    join_node = JoinToBaseOutputNode[DataSourceDataSet](
        parent_node=bookings_filtered,
        join_targets=[
            JoinDescription(
                join_node=listings_filtered,
                join_on_identifier=LinklessIdentifierSpec.from_element_name("listing"),
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            )
        ],
    )

    bookings_aggregated = AggregateMeasuresNode[DataSourceDataSet](parent_node=join_node)

    cost_function = DefaultCostFunction[DataSourceDataSet]()
    cost = cost_function.calculate_cost(bookings_aggregated)

    assert cost == DefaultCost(num_joins=1, num_aggregations=1)



from __future__ import annotations

from typing import Optional

from metricflow.model.objects.base import HashableBaseModel
from metricflow.specs import DimensionReference, TimeDimensionReference
from metricflow.time.time_granularity import TimeGranularity
from metricflow.object_utils import ExtendedEnum

ISO8601_FMT = "YYYY-MM-DD"


class DimensionType(ExtendedEnum):
    """Determines types of values expected of dimensions."""

    CATEGORICAL = "categorical"
    TIME = "time"

    def is_time_type(self) -> bool:
        """Checks if this type of dimension is a time type"""
        return self in [DimensionType.TIME]


class DimensionTypeParams(HashableBaseModel):
    """Dimension type params add additional context to some types (time) of dimensions"""

    is_primary: bool = False
    # For legacy support. This is not used.
    time_format: str = ISO8601_FMT
    time_granularity: TimeGranularity


class Dimension(HashableBaseModel):
    """Describes a dimension"""

    name: str
    description: Optional[str]
    type: DimensionType
    is_partition: bool = False
    type_params: Optional[DimensionTypeParams]
    expr: Optional[str] = None

    @property
    def is_primary_time(self) -> bool:  # noqa: D
        if self.type == DimensionType.TIME and self.type_params is not None:
            return self.type_params.is_primary

        return False

    @property
    def reference(self) -> DimensionReference:  # noqa: D
        return DimensionReference(element_name=self.name)

    @property
    def time_dimension_reference(self) -> TimeDimensionReference:  # noqa: D
        assert self.type == DimensionType.TIME, f"Got type as {self.type} instead of {DimensionType.TIME}"
        return TimeDimensionReference(element_name=self.name)



import datetime as dt

from dateutil.parser import parse
from typing import Optional

from metricflow.configuration.constants import CONFIG_MODEL_PATH
from metricflow.configuration.yaml_handler import YamlFileHandler
from metricflow.errors.errors import ModelCreationException
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.parsing.dir_to_model import ModelBuildResult, parse_directory_of_yaml_files_to_model
from metricflow.sql_clients.common_client import not_empty


def path_to_models(handler: YamlFileHandler) -> str:
    """Given a YamlFileHandler, return the path to the YAML model config files"""
    return not_empty(handler.get_value(CONFIG_MODEL_PATH), CONFIG_MODEL_PATH, handler.url)


def model_build_result_from_config(
    handler: YamlFileHandler, raise_issues_as_exceptions: bool = True
) -> ModelBuildResult:
    """Given a yaml file, creates a ModelBuildResult.

    Args:
        handler: a file handler for loading the configs from
        raise_issues_as_exceptions: determines if issues should be raised, or returned as issues

    Returns:
        ModelBuildResult that contains the UserConfigureModel and any associated ValidationIssues
    """
    models_path = path_to_models(handler=handler)
    try:
        return parse_directory_of_yaml_files_to_model(
            models_path, raise_issues_as_exceptions=raise_issues_as_exceptions
        )
    except Exception as e:
        raise ModelCreationException from e


def build_user_configured_model_from_config(handler: YamlFileHandler) -> UserConfiguredModel:
    """Given a yaml file, create a UserConfiguredModel."""
    return model_build_result_from_config(handler=handler).model


def convert_to_datetime(datetime_str: Optional[str]) -> Optional[dt.datetime]:
    """Callback to convert string to datetime given as an iso8601 timestamp."""
    if datetime_str is None:
        return None

    try:
        return parse(datetime_str)
    except Exception:
        raise ValueError(f"'{datetime_str}' is not a valid iso8601 timestamp")



import logging

import pytest

from metricflow.telemetry.models import TelemetryLevel
from metricflow.telemetry.reporter import TelemetryReporter, log_call

logger = logging.getLogger(__name__)


@pytest.fixture
def telemetry_reporter() -> TelemetryReporter:  # noqa: D
    reporter = TelemetryReporter(report_levels_higher_or_equal_to=TelemetryLevel.USAGE)
    reporter.add_python_log_handler()
    reporter.add_test_handler()
    return reporter


def test_function_call(telemetry_reporter: TelemetryReporter) -> None:  # noqa: D
    @log_call(telemetry_reporter=telemetry_reporter, module_name=__name__)
    def test_function() -> str:
        return "foo"

    test_function()

    start_event = telemetry_reporter.test_handler.payloads[0].function_start_events[0]
    assert start_event.module_name == "metricflow.test.telemetry.test_telemetry"
    assert start_event.function_name == "test_function"

    end_event = telemetry_reporter.test_handler.payloads[1].function_end_events[0]
    assert end_event.module_name == "metricflow.test.telemetry.test_telemetry"
    assert end_event.function_name == "test_function"
    assert not end_event.exception_trace
    assert end_event.runtime > 0


def test_function_exception(telemetry_reporter: TelemetryReporter) -> None:  # noqa: D
    with pytest.raises(ValueError):

        @log_call(telemetry_reporter=telemetry_reporter, module_name=__name__)
        def test_function() -> str:
            raise ValueError("foo")

        test_function()

    start_event = telemetry_reporter.test_handler.payloads[0].function_start_events[0]
    assert start_event.module_name == "metricflow.test.telemetry.test_telemetry"
    assert start_event.function_name == "test_function"

    end_event = telemetry_reporter.test_handler.payloads[1].function_end_events[0]
    assert end_event.module_name == "metricflow.test.telemetry.test_telemetry"
    assert end_event.function_name == "test_function"
    assert end_event.exception_trace
    assert end_event.exception_trace.find("Traceback (most recent call last):") != -1
    assert end_event.exception_trace.find("ValueError: foo") != -1
    assert end_event.runtime > 0


def test_telemetry_off() -> None:  # noqa: D
    reporter = TelemetryReporter(report_levels_higher_or_equal_to=TelemetryLevel.OFF)
    reporter.add_python_log_handler()
    reporter.add_test_handler()

    @log_call(telemetry_reporter=reporter, module_name=__name__)
    def test_function() -> str:
        return "foo"

    test_function()

    with pytest.raises(ValueError):

        @log_call(telemetry_reporter=reporter, module_name=__name__)
        def test_exception_function() -> str:
            raise ValueError("foo")

        test_exception_function()

    assert len(reporter.test_handler.payloads) == 0


@pytest.mark.skip("Sends events to Rudderstack")
def test_rudderstack_logging() -> None:  # noqa: D
    reporter = TelemetryReporter(report_levels_higher_or_equal_to=TelemetryLevel.USAGE)
    reporter.add_python_log_handler()
    reporter.add_rudderstack_handler()

    @log_call(telemetry_reporter=reporter, module_name=__name__)
    def test_function() -> str:
        return "foo"

    test_function()

    with pytest.raises(ValueError):

        @log_call(telemetry_reporter=reporter, module_name=__name__)
        def test_exception_function() -> str:
            raise ValueError("foo")

        test_exception_function()



import re
from jsonschema import Draft7Validator, ValidationError
from jsonschema._utils import extras_msg
from jsonschema.validators import extend


def custom_find_additional_properties(instance, schema):  # type: ignore[no-untyped-def]
    """Return the set of additional properties for the given ``instance``.

    NOTE: This is a modified copy of the ``find_additional_properties`` method
    defined in jsonschema/_utils.py jsonschema 3.2.0 found here
    https://github.com/python-jsonschema/jsonschema/blob/2f734a7ce1395ac4ff2d394583fd46a2e8833a9b/jsonschema/_utils.py#L84

    Weeds out properties that should have been validated by ``properties`` and
    / or ``patternProperties``. Additionally, completely ignores any properties
    matching ``^__(.+)__$``

    Assumes ``instance`` is dict-like already.
    """

    properties = schema.get("properties", {})
    schema_patterns = schema.get("patternProperties", {})
    # we ignore all things that match "^__(.+)__$" in all objects
    schema_patterns["^__(.+)__$"] = {}
    patterns = "|".join(schema_patterns)
    for property in instance:
        if property not in properties:
            if patterns and re.search(patterns, property):
                continue
            yield property


def customAdditionalProperties(validator, aP, instance, schema):  # type: ignore[no-untyped-def]
    """Validator for checking if a schema has additionalProperties when it shouldn't

    NOTE: This is a modified copy of the ``additionalProperties`` method
    defined in jsonschema/_validators.py of jsonschema 3.2.0 found here
    https://github.com/python-jsonschema/jsonschema/blob/2f734a7ce1395ac4ff2d394583fd46a2e8833a9b/jsonschema/_validators.py#L41
    """
    if not validator.is_type(instance, "object"):
        return

    extras = set(custom_find_additional_properties(instance, schema))

    if validator.is_type(aP, "object"):
        for extra in extras:
            for error in validator.descend(instance[extra], aP, path=extra):
                yield error
    elif not aP and extras:
        if "patternProperties" in schema:
            patterns = sorted(schema["patternProperties"])
            if len(extras) == 1:
                verb = "does"
            else:
                verb = "do"
            error = "%s %s not match any of the regexes: %s" % (
                ", ".join(map(repr, sorted(extras))),
                verb,
                ", ".join(map(repr, patterns)),
            )
            yield ValidationError(error)
        else:
            error = "Additional properties are not allowed (%s %s unexpected)"
            yield ValidationError(error % extras_msg(extras))


# Extend takes a given validator, and overrides/adds the specified validators for the validator
# Thus here we are overriding Draft7Validator's `additionalProperties` validator
SchemaValidator = extend(validator=Draft7Validator, validators={"additionalProperties": customAdditionalProperties})



from typing import List

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    ModelValidationRule,
    ValidationError,
    ValidationIssueType,
    validate_safely,
)


class NonEmptyRule(ModelValidationRule):
    """Check if the model contains data sources and metrics."""

    @staticmethod
    @validate_safely(whats_being_done="checking that the model has data sources")
    def _check_model_has_data_sources(model: UserConfiguredModel) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []
        if not model.data_sources:
            issues.append(
                ValidationError(
                    message="No data sources present in the model.",
                )
            )
        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking that the model has metrics")
    def _check_model_has_metrics(model: UserConfiguredModel) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        # If we are going to generate measure proxy metrics that is sufficient as well
        create_measure_proxy_metrics = False
        for data_source in model.data_sources:
            for measure in data_source.measures:
                if measure.create_metric is True:
                    create_measure_proxy_metrics = True
                    break

        if not model.metrics and not create_measure_proxy_metrics:
            issues.append(
                ValidationError(
                    message="No metrics present in the model.",
                )
            )
        return issues

    @staticmethod
    @validate_safely("running model validation rule ensuring metrics and data sources are defined")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []
        issues += NonEmptyRule._check_model_has_data_sources(model=model)
        issues += NonEmptyRule._check_model_has_metrics(model=model)
        return issues



from __future__ import annotations

import logging
import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Tuple, Sequence

from metricflow.sql.render.expr_renderer import DefaultSqlExpressionRenderer, SqlExpressionRenderer
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql.sql_plan import (
    SqlQueryPlanNodeVisitor,
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlQueryPlanNode,
    SqlQueryPlan,
    SqlSelectQueryFromClauseNode,
    SqlSelectColumn,
    SqlJoinDescription,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SqlPlanRenderResult:  # noqa: D
    # The SQL string that could be run.
    sql: str
    # The execution parameters that should be specified along with the SQL str to execute()
    execution_parameters: SqlBindParameters


class SqlQueryPlanRenderer(SqlQueryPlanNodeVisitor[SqlPlanRenderResult], ABC):
    """Renders SQL plans to a string."""

    def _render_node(self, node: SqlQueryPlanNode) -> SqlPlanRenderResult:  # noqa: D
        return node.accept(self)

    def render_sql_query_plan(self, sql_query_plan: SqlQueryPlan) -> SqlPlanRenderResult:  # noqa: D
        return self._render_node(sql_query_plan.render_node)

    @property
    @abstractmethod
    def expr_renderer(self) -> SqlExpressionRenderer:  # noqa: D
        """Return the renderer that this uses to render expressions."""
        pass


@dataclass
class StringJoinDescription:
    """Helper class to store the join clause as a string. Useful to avoid logic in the Jinja template."""

    join_source_str: str
    join_source_is_a_table: bool
    join_source_alias: str
    on_condition_str: str


class DefaultSqlQueryPlanRenderer(SqlQueryPlanRenderer):
    """Renders an SQL plan following ANSI SQL"""

    # The renderer that is used to render the SQL expressions.
    EXPR_RENDERER = DefaultSqlExpressionRenderer()
    # The amount to indent when formatting SQL
    INDENT = "  "

    def _render_select_columns_section(
        self,
        select_columns: Sequence[SqlSelectColumn],
        num_parents: int,
    ) -> Tuple[str, SqlBindParameters]:
        """Convert the select columns into a "SELECT" section.

        e.g.
        SELECT
          1 AS bookings
          , listing_id AS listing
          ...

        Returns a tuple of the "SELECT" section as a string and the associated execution parameters.
        """
        params = SqlBindParameters()
        select_section_lines = ["SELECT"]
        first_column = True
        for select_column in select_columns:
            expr_rendered = self.EXPR_RENDERER.render_sql_expr(select_column.expr)
            # Merge all execution parameters together. Similar pattern follows below.
            params.update(expr_rendered.execution_parameters)

            column_select_str = f"{expr_rendered.sql} AS {select_column.column_alias}"

            # For cases where the alias is the same as column like "src.foo AS foo", just render it as "src.foo"
            # SQLite throws an ambiguous column error with something like
            #
            # SELECT a.ds
            # FROM fct_bookings a
            # JOIN dim_users b
            # on a.ds = b.ds
            #
            # so don't do this if it has joins.
            if num_parents <= 1 and select_column.expr.as_column_reference_expression:
                column_reference = select_column.expr.as_column_reference_expression.col_ref
                if column_reference.column_name == select_column.column_alias:
                    column_select_str = expr_rendered.sql

            if first_column:
                first_column = False
                select_section_lines.append(f"{self.INDENT}{column_select_str}")
            else:
                select_section_lines.append(f"{self.INDENT}, {column_select_str}")

        return "\n".join(select_section_lines), params

    def _render_from_section(
        self, from_source: SqlQueryPlanNode, from_source_alias: str
    ) -> Tuple[str, SqlBindParameters]:
        """Convert the node into a "FROM" section.

        e.g.
        FROM (
          SELECT
            1 AS bookings
            ...
        )

        Returns a tuple of the "FROM" section as a string and the associated execution parameters.
        """
        from_render_result = self._render_node(from_source)

        from_section_lines = []
        if from_source.is_table:
            from_section_lines.append(f"FROM {from_render_result.sql} {from_source_alias}")
        else:
            from_section_lines.append("FROM (")
            from_section_lines.append(textwrap.indent(from_render_result.sql, prefix=self.INDENT))
            from_section_lines.append(f") {from_source_alias}")
        from_section = "\n".join(from_section_lines)

        return from_section, from_render_result.execution_parameters

    def _render_joins_section(self, join_descriptions: Sequence[SqlJoinDescription]) -> Tuple[str, SqlBindParameters]:
        """Convert the join descriptions into a "JOIN" section.

        e.g.
        JOIN (
          SELECT...
        ) right
        ON
          left.ds = right.ds

        Returns a tuple of the "JOIN" section as a string and the associated execution parameters.
        """
        params = SqlBindParameters()
        join_section_lines = []
        for join_description in join_descriptions:
            # Render the source for the join
            right_source_rendered = self._render_node(join_description.right_source)
            params.update(right_source_rendered.execution_parameters)

            # Render the on condition for the join
            on_condition_rendered = self.EXPR_RENDERER.render_sql_expr(join_description.on_condition)
            params.update(on_condition_rendered.execution_parameters)

            if join_description.right_source.is_table:
                join_section_lines.append(join_description.join_type.value)
                join_section_lines.append(
                    textwrap.indent(
                        f"{right_source_rendered.sql} {join_description.right_source_alias}", prefix=self.INDENT
                    )
                )
            else:
                join_section_lines.append(f"{join_description.join_type.value} (")
                join_section_lines.append(textwrap.indent(right_source_rendered.sql, prefix=self.INDENT))
                join_section_lines.append(f") {join_description.right_source_alias}")
            join_section_lines.append("ON")
            join_section_lines.append(textwrap.indent(on_condition_rendered.sql, prefix=self.INDENT))

        return "\n".join(join_section_lines), params

    def _render_group_by_section(self, group_by_columns: Sequence[SqlSelectColumn]) -> Tuple[str, SqlBindParameters]:
        """Convert the group by columns into a "GROUP BY" section.

        e.g.
        GROUP BY
          a.ds

        Returns a tuple of the "GROUP BY" section as a string and the associated execution parameters.
        """
        group_by_section_lines: List[str] = []
        params = SqlBindParameters()
        first = True
        for group_by_column in group_by_columns:
            group_by_expr_rendered = self.EXPR_RENDERER.render_sql_expr(group_by_column.expr)
            params.update(group_by_expr_rendered.execution_parameters)
            if first:
                first = False
                group_by_section_lines.append("GROUP BY")
                group_by_section_lines.append(textwrap.indent(group_by_expr_rendered.sql, prefix=self.INDENT))
            else:
                group_by_section_lines.append(textwrap.indent(f", {group_by_expr_rendered.sql}", prefix=self.INDENT))

        return "\n".join(group_by_section_lines), params

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlPlanRenderResult:  # noqa: D
        # Keep track of all execution parameters for all expressions
        combined_params = SqlBindParameters()

        # Render description section
        description_section = "\n".join([f"-- {x}" for x in node.description.split("\n")])

        # Render "SELECT" column section
        select_section, select_params = self._render_select_columns_section(node.select_columns, len(node.parent_nodes))
        combined_params.update(select_params)

        # Render "FROM" section
        from_section, from_params = self._render_from_section(node.from_source, node.from_source_alias)
        combined_params.update(from_params)

        # Render "JOIN" section
        join_section, join_params = self._render_joins_section(node.join_descs)
        combined_params.update(join_params)

        # Render "GROUP BY" section
        group_by_section, group_by_params = self._render_group_by_section(node.group_bys)
        combined_params.update(group_by_params)

        # Render "WHERE" section
        where_section = None
        if node.where:
            where_render_result = self.EXPR_RENDERER.render_sql_expr(node.where)
            combined_params.update(where_render_result.execution_parameters)
            where_section = f"WHERE {where_render_result.sql}"

        # Render "ORDER BY" section
        order_by_section = None
        if node.order_bys:
            order_by_items: List[str] = []
            for order_by in node.order_bys:
                order_by_render_result = self.EXPR_RENDERER.render_sql_expr(order_by.expr)
                order_by_items.append(order_by_render_result.sql + (" DESC" if order_by.desc else ""))
                combined_params.update(order_by_render_result.execution_parameters)

            order_by_section = "ORDER BY " + ", ".join(order_by_items)

        # Render "LIMIT" section
        limit_section = None
        if node.limit:
            limit_section = f"LIMIT {node.limit}"

        # Combine the sections into a single string.
        sections_to_render = []

        if description_section:
            sections_to_render.append(description_section)

        sections_to_render.append(select_section)
        sections_to_render.append(from_section)

        if join_section:
            sections_to_render.append(join_section)

        if where_section:
            sections_to_render.append(where_section)

        if group_by_section:
            sections_to_render.append(group_by_section)

        if order_by_section:
            sections_to_render.append(order_by_section)

        if limit_section:
            sections_to_render.append(limit_section)

        return SqlPlanRenderResult(
            sql="\n".join(sections_to_render),
            execution_parameters=combined_params,
        )

    def visit_table_from_clause_node(self, node: SqlTableFromClauseNode) -> SqlPlanRenderResult:  # noqa: D
        return SqlPlanRenderResult(
            sql=node.sql_table.sql,
            execution_parameters=SqlBindParameters(),
        )

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlPlanRenderResult:  # noqa: D
        return SqlPlanRenderResult(
            sql=node.select_query.rstrip(),
            execution_parameters=SqlBindParameters(),
        )

    @property
    def expr_renderer(self) -> SqlExpressionRenderer:  # noqa :D
        return self.EXPR_RENDERER



from typing import Set

from metricflow.specs import InstanceSpecSetTransform, InstanceSpecSet


class ToElementNameSet(InstanceSpecSetTransform[Set[str]]):
    """Gets all element names for all specs in the set."""

    def transform(self, spec_set: InstanceSpecSet) -> Set[str]:  # noqa: D
        return (
            {x.element_name for x in spec_set.metric_specs}
            .union({x.element_name for x in spec_set.measure_specs})
            .union({x.element_name for x in spec_set.dimension_specs})
            .union({x.element_name for x in spec_set.time_dimension_specs})
            .union({x.element_name for x in spec_set.identifier_specs})
        )



from typing import Dict, Any

from jsonschema import RefResolver
from metricflow.model.parsing.schema_validator import SchemaValidator

from metricflow.model.parsing.schemas import (
    metric_schema,
    data_source_schema,
    materialization_schema,
    derived_group_by_element_schema,
    metric_input_measure_schema,
    metric_type_params_schema,
    identifier_schema,
    measure_schema,
    dimension_schema,
    dimension_type_params_schema,
    mutability_schema,
    mutability_type_params_schema,
    composite_sub_identifier_schema,
    materialization_destination_schema,
)


def add_transform_metadata_fields_to_spec(spec: Dict[str, Any]) -> None:  # type: ignore[misc]
    """Adds transform metadata fields a spec"""
    properties = spec["properties"]
    transform_metadata_fields = {
        "description": {"type": "string"},
        "owners": {
            "type": "array",
            "items": {"type": "string"},
        },
        "display_name": {"type": "string"},
        "tier": {"type": ["string", "integer"]},
    }

    for k, v in transform_metadata_fields.items():
        if k in properties:
            raise RuntimeError(f"Spec with id: {spec['$id']} contains transform metadata field: {k}")

        properties[k] = v


def add_locked_metadata_to_spec(spec: Dict[str, Any]) -> None:  # type: ignore[misc]
    """Adds locked metadata field to a spec"""
    properties = spec["properties"]
    transform_metadata_fields = {
        "locked_metadata": {"$ref": "locked_metadata"},
    }

    for k, v in transform_metadata_fields.items():
        if k in properties:
            raise RuntimeError(f"Spec with id: {spec['$id']} contains transform metadata field: {k}")

        properties[k] = v


# Sub-object schemas
locked_metadata_schema = {
    "$id": "locked_metadata",
    "type": "object",
    "properties": {
        "value_format": {"type": "string"},
        "description": {"type": "string"},
        "display_name": {"type": "string"},
        "tier": {"type": ["string", "integer"]},
        "increase_is_good": {"type": "boolean"},
    },
    "additionalProperties": False,
}


# Add transform metadata fields to top level objects (metric, data source, derived identifier)
add_transform_metadata_fields_to_spec(dimension_schema)
add_transform_metadata_fields_to_spec(measure_schema)
add_transform_metadata_fields_to_spec(identifier_schema)

add_transform_metadata_fields_to_spec(metric_schema)
add_locked_metadata_to_spec(metric_schema)

add_transform_metadata_fields_to_spec(data_source_schema)
add_transform_metadata_fields_to_spec(materialization_schema)
add_transform_metadata_fields_to_spec(derived_group_by_element_schema)


schema_store = {
    # Top level schemas
    metric_schema["$id"]: metric_schema,
    data_source_schema["$id"]: data_source_schema,
    derived_group_by_element_schema["$id"]: derived_group_by_element_schema,
    materialization_schema["$id"]: materialization_schema,
    # Sub-object schemas
    metric_input_measure_schema["$id"]: metric_input_measure_schema,
    metric_type_params_schema["$id"]: metric_type_params_schema,
    locked_metadata_schema["$id"]: locked_metadata_schema,
    identifier_schema["$id"]: identifier_schema,
    measure_schema["$id"]: measure_schema,
    dimension_schema["$id"]: dimension_schema,
    dimension_type_params_schema["$id"]: dimension_type_params_schema,
    mutability_schema["$id"]: mutability_schema,
    mutability_type_params_schema["$id"]: mutability_type_params_schema,
    composite_sub_identifier_schema["$id"]: composite_sub_identifier_schema,
    materialization_destination_schema["$id"]: materialization_destination_schema,
}

resolver = RefResolver.from_schema(schema=metric_schema, store=schema_store)
data_source_validator = SchemaValidator(data_source_schema, resolver=resolver)
derived_group_by_element_validator = SchemaValidator(derived_group_by_element_schema, resolver=resolver)
materialization_validator = SchemaValidator(materialization_schema, resolver=resolver)
metric_validator = SchemaValidator(metric_schema, resolver=resolver)



import copy
import logging

from typing import Sequence

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.transformations.agg_time_dimension import SetMeasureAggregationTimeDimensionRule
from metricflow.model.transformations.boolean_measure import BooleanMeasureAggregationRule
from metricflow.model.transformations.identifiers import CompositeIdentifierExpressionRule
from metricflow.model.transformations.names import LowerCaseNamesRule
from metricflow.model.transformations.proxy_measure import CreateProxyMeasureRule
from metricflow.model.transformations.transform_rule import ModelTransformRule

logger = logging.getLogger(__name__)


class ModelTransformer:
    """Helps to make transformations to a model for convenience.

    Generally used to make it more convenient for the user to develop their model.
    """

    DEFAULT_PRE_VALIDATION_RULES: Sequence[ModelTransformRule] = (
        LowerCaseNamesRule(),
        SetMeasureAggregationTimeDimensionRule(),
    )

    DEFAULT_POST_VALIDATION_RULES: Sequence[ModelTransformRule] = (
        CreateProxyMeasureRule(),
        BooleanMeasureAggregationRule(),
        CompositeIdentifierExpressionRule(),
    )

    @staticmethod
    def pre_validation_transform_model(
        model: UserConfiguredModel, rules: Sequence[ModelTransformRule] = DEFAULT_PRE_VALIDATION_RULES
    ) -> UserConfiguredModel:
        """Transform a model according to configured rules before validations are run."""
        model_copy = copy.deepcopy(model)

        for transform_rule in rules:
            model_copy = transform_rule.transform_model(model_copy)

        return model_copy

    @staticmethod
    def post_validation_transform_model(
        model: UserConfiguredModel, rules: Sequence[ModelTransformRule] = DEFAULT_POST_VALIDATION_RULES
    ) -> UserConfiguredModel:
        """Transform a model according to configured rules after validations are run."""
        model_copy = copy.deepcopy(model)
        for transform_rule in rules:
            model_copy = transform_rule.transform_model(model_copy)

        return model_copy



from __future__ import annotations

import datetime
import logging
import threading
from dataclasses import dataclass
from typing import List, Tuple

import pandas as pd

from metricflow.protocols.sql_client import SqlClient
from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.dataflow.sql_table import SqlTable
from metricflow.time.time_constants import ISO8601_PYTHON_FORMAT
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class TimeSpineSource:
    """Defines a source table containing all timestamps to use for computing cumulative metrics."""

    schema_name: str
    table_name: str = "mf_time_spine"
    # Name of the column in the table that contains the dates.
    time_column_name: str = "ds"
    # The time granularity of the dates in the spine table.
    time_column_granularity: TimeGranularity = TimeGranularity.DAY

    @property
    def spine_table(self) -> SqlTable:
        """Table containing all dates"""
        return SqlTable(schema_name=self.schema_name, table_name=self.table_name)


class TimeSpineTableBuilder:
    """Helps to build the time spine table based on the definition in a TimeSpineSource."""

    def __init__(  # noqa: D
        self,
        time_spine_source: TimeSpineSource,
        sql_client: SqlClient,
    ) -> None:
        self._time_spine_source = time_spine_source
        self._sql_client = sql_client
        self._verified_spine_table_exists = False
        self._create_table_lock = threading.Lock()

    @property
    def time_spine_source(self) -> TimeSpineSource:  # noqa: D
        return self._time_spine_source

    def create_if_necessary(self) -> None:  # noqa: D
        """Creates the spine table if it doesn't already exist."""
        logger.info("Waiting to get the lock for the time spine table")
        with self._create_table_lock:
            spine_table = self.time_spine_source.spine_table
            logger.info("Got the lock for the time spine table")
            if self._verified_spine_table_exists:
                logger.info(f"Previously verified that the spine table {spine_table.sql} exists.")
                return
            logger.info(f"Checking if the spine table {spine_table.sql} exists")
            if self._sql_client.table_exists(spine_table):
                logger.info(f"Spine table {spine_table.sql} exists")
                self._verified_spine_table_exists = True
                return
            logger.info(f"Spine table {spine_table.sql} does not exist")
            start_date = TimeRangeConstraint.ALL_TIME_BEGIN()
            end_date = TimeRangeConstraint.ALL_TIME_END()

            current_date = start_date
            # Using a union type throws a type error for some reason, so going with this approach
            date_spine_table_datetime_data: List[Tuple[datetime.datetime]] = []
            date_spine_table_str_data: List[Tuple[str]] = []

            if self.time_spine_source.time_column_granularity != TimeGranularity.DAY:
                raise RuntimeError(
                    f"A time spine source with a granularity {self.time_spine_source.time_column_granularity} is not "
                    f"yet supported."
                )

            if self._sql_client.sql_engine_attributes.timestamp_type_supported:
                while current_date <= end_date:
                    date_spine_table_datetime_data.append((current_date,))
                    current_date = current_date + datetime.timedelta(days=1)
            else:
                while current_date <= end_date:
                    date_spine_table_str_data.append((current_date.strftime(ISO8601_PYTHON_FORMAT),))
                    current_date = current_date + datetime.timedelta(days=1)

            self._sql_client.drop_table(spine_table)
            num_rows = (
                len(date_spine_table_datetime_data)
                if date_spine_table_datetime_data
                else len(date_spine_table_str_data)
            )

            logger.info(f"Creating date spine table {spine_table.sql} with {num_rows} rows")
            self._sql_client.create_table_from_dataframe(
                sql_table=spine_table,
                df=pd.DataFrame(
                    columns=[self._time_spine_source.time_column_name],
                    data=date_spine_table_datetime_data or date_spine_table_str_data,
                ),
                chunk_size=1000,
            )
            logger.info(f"Created date spine table {spine_table.sql}")
            self._verified_spine_table_exists = True



from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantics.data_source_container import PydanticDataSourceContainer
from metricflow.model.semantics.semantic_containers import DataSourceSemantics, MetricSemantics


class SemanticModel:
    """Adds semantics information to the user configured model."""

    def __init__(self, user_configured_model: UserConfiguredModel) -> None:  # noqa: D
        self._user_configured_model = user_configured_model
        self._data_source_semantics = DataSourceSemantics(
            user_configured_model, PydanticDataSourceContainer(user_configured_model.data_sources)
        )
        self._metric_semantics = MetricSemantics(self._user_configured_model, self._data_source_semantics)

    @property
    def user_configured_model(self) -> UserConfiguredModel:  # noqa: D
        return self._user_configured_model

    @property
    def data_source_semantics(self) -> DataSourceSemantics:  # noqa: D
        return self._data_source_semantics

    @property
    def metric_semantics(self) -> MetricSemantics:  # noqa: D
        return self._metric_semantics



"""These spec classes generally describe something that we want or have.

* Not too many comments here since they seem mostly self-explanatory.
* To see whether a spec matches something that already exists, there could be a method that allows you to match a spec
  to another spec or relevant object.
* The match() method will enable sub-classes (may require some restructuring) to use specs to request things like,
  metrics named "sales*".
"""

from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple, TypeVar, Generic

from metricflow.column_assoc import ColumnAssociation
from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.time.time_granularity import TimeGranularity
from metricflow.model.objects.base import FrozenBaseModel
from metricflow.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow.sql.sql_bind_parameters import SqlBindParameters


class ColumnAssociationResolver(ABC):
    """Get the default column associations for an element instance.

    This is used for naming columns in an SQL query consistently. For example, dimensions with links are
    named like <identifier link>__<dimension name> e.g. user_id__country, and time dimensions at a different time
    granularity are named <time dimension>__<time granularity> e.g. ds__month. Having a central place to name them will
    make it easier to change this later on. Names generated need to be unique within a query.

    It's also important to maintain this format because customers write constraints in SQL assuming this. This
    allows us to stick the constraint in as WHERE clauses without having to parse the constraint SQL.

    TODO: Updates are needed for time granularity in time dimensions, ToT for metrics.

    The resolve* methods should return the column associations / column names that it should use in queries for the given
    spec.
    """

    @abstractmethod
    def resolve_metric_spec(self, metric_spec: MetricSpec) -> ColumnAssociation:  # noqa: D
        pass

    @abstractmethod
    def resolve_measure_spec(self, measure_spec: MeasureSpec) -> ColumnAssociation:  # noqa: D
        pass

    @abstractmethod
    def resolve_dimension_spec(self, dimension_spec: DimensionSpec) -> ColumnAssociation:  # noqa: D
        pass

    @abstractmethod
    def resolve_time_dimension_spec(self, time_dimension_spec: TimeDimensionSpec) -> ColumnAssociation:  # noqa: D
        pass

    @abstractmethod
    def resolve_identifier_spec(self, identifier_spec: IdentifierSpec) -> Tuple[ColumnAssociation, ...]:  # noqa: D
        pass


class InstanceSpec(ABC, FrozenBaseModel):
    """A specification for an instance of a metric definition object.

    An instance is different from the definition object in that it correlates to columns in the data flow and can be in
    different states. e.g. a time dimension at a different time granularity.

    Added "type: ignore" as there is an issue with abstract methods in frozen dataclasses:
    https://github.com/python/mypy/issues/5374
    """

    """Name of the dimension or identifier in the data source."""
    element_name: str

    @abstractmethod
    def column_associations(self, resolver: ColumnAssociationResolver) -> Tuple[ColumnAssociation, ...]:
        """Figures out what columns in an SQL query that this spec should be associated with given the resolver.

        Debating whether this should be an abstract method, or whether it could just live in the different specs to
        allow for different signatures.
        """
        pass

    @staticmethod
    def merge(*specs: Sequence[InstanceSpec]) -> List[InstanceSpec]:
        """Merge all specs into a single list."""
        result: List[InstanceSpec] = []
        for spec in specs:
            result.extend(spec)
        return result

    @property
    @abstractmethod
    def qualified_name(self) -> str:
        """Return the qualified name of this spec. e.g. "user_id__country"."""
        pass


class LinkableInstanceSpec(InstanceSpec, ABC):
    """Generally a dimension or identifier that may be specified using identifier links.

    For example, user_id__country -> LinkableElementSpec(element_name="country", identifier_links=["user_id"]

    See InstanceSpec for the reason behind "type: ignore"
    """

    """A list representing the join path of identifiers to get to this element."""
    identifier_links: Tuple[LinklessIdentifierSpec, ...]

    def __post_init__(self) -> None:  # noqa: D
        for identifier in self.identifier_links:
            if len(identifier.identifier_links) > 0:
                raise RuntimeError("Identifier links may not have their own identifier links")

    @abstractmethod
    def without_first_identifier_link(self) -> LinkableInstanceSpec:
        """e.g. user_id__device_id__platform -> device_id__platform"""
        pass

    @abstractmethod
    def without_identifier_links(self) -> LinkableInstanceSpec:  # noqa: D
        """e.g. user_id__device_id__platform -> platform"""
        pass

    @staticmethod
    def merge_linkable_specs(*specs: Sequence[LinkableInstanceSpec]) -> List[LinkableInstanceSpec]:
        """Merge all specs into a single list."""
        result: List[LinkableInstanceSpec] = []
        for spec in specs:
            result.extend(spec)
        return result

    @property
    def qualified_name(self) -> str:
        """Return the qualified name of this spec. e.g. "user_id__country"."""
        return StructuredLinkableSpecName(
            identifier_link_names=tuple(x.element_name for x in self.identifier_links), element_name=self.element_name
        ).qualified_name


class IdentifierSpec(LinkableInstanceSpec):  # noqa: D
    def column_associations(self, resolver: ColumnAssociationResolver) -> Tuple[ColumnAssociation, ...]:  # noqa: D
        return resolver.resolve_identifier_spec(self)

    def without_first_identifier_link(self) -> LinkableInstanceSpec:  # noqa: D
        assert len(self.identifier_links) > 0, f"Spec does not have any identifier links: {self}"
        return IdentifierSpec(element_name=self.element_name, identifier_links=self.identifier_links[1:])

    def without_identifier_links(self) -> LinklessIdentifierSpec:  # noqa: D
        return LinklessIdentifierSpec.from_element_name(self.element_name)

    @property
    def as_linkless_prefix(self) -> Tuple[LinklessIdentifierSpec, ...]:
        """Creates tuple of linkless identifiers that could be included in the identifier_links of another spec

        eg as a prefix to a DimensionSpec's identifier links to when a join is occurring via this identifier
        """
        return (LinklessIdentifierSpec.from_element_name(self.element_name),) + self.identifier_links

    @staticmethod
    def from_name(name: str) -> IdentifierSpec:  # noqa: D
        structured_name = StructuredLinkableSpecName.from_name(name)
        return IdentifierSpec(
            identifier_links=tuple(
                LinklessIdentifierSpec.from_element_name(idl) for idl in structured_name.identifier_link_names
            ),
            element_name=structured_name.element_name,
        )


class LinklessIdentifierSpec(IdentifierSpec):
    """Similar to IdentifierSpec, but requires that it doesn't have identifier links."""

    @staticmethod
    def from_element_name(element_name: str) -> LinklessIdentifierSpec:  # noqa: D
        return LinklessIdentifierSpec(element_name=element_name, identifier_links=())

    def __post_init__(self) -> None:  # noqa: D
        if len(self.identifier_links) > 0:
            raise RuntimeError(f"{self.__class__.__name__} shouldn't have identifier links. Got: {self}")


IdentifierSpec.update_forward_refs()
LinklessIdentifierSpec.update_forward_refs()


class DimensionSpec(LinkableInstanceSpec):  # noqa: D
    def column_associations(self, resolver: ColumnAssociationResolver) -> Tuple[ColumnAssociation, ...]:  # noqa: D
        return (resolver.resolve_dimension_spec(self),)

    def without_first_identifier_link(self) -> LinkableInstanceSpec:  # noqa: D
        assert len(self.identifier_links) > 0, f"Spec does not have any identifier links: {self}"
        return DimensionSpec(element_name=self.element_name, identifier_links=self.identifier_links[1:])

    def without_identifier_links(self) -> LinkableInstanceSpec:  # noqa: D
        return DimensionSpec(element_name=self.element_name, identifier_links=())

    @staticmethod
    def from_linkable(spec: LinkableInstanceSpec) -> DimensionSpec:  # noqa: D
        return DimensionSpec(element_name=spec.element_name, identifier_links=spec.identifier_links)

    @staticmethod
    def from_name(name: str) -> DimensionSpec:
        """Construct from a name e.g. listing__ds__month."""
        parsed_name = StructuredLinkableSpecName.from_name(name)
        return DimensionSpec(
            identifier_links=tuple(
                [LinklessIdentifierSpec.from_element_name(idl) for idl in parsed_name.identifier_link_names]
            ),
            element_name=parsed_name.element_name,
        )

    @property
    def reference(self) -> DimensionReference:  # noqa: D
        return DimensionReference(element_name=self.element_name)


DimensionSpec.update_forward_refs()

DEFAULT_TIME_GRANULARITY = TimeGranularity.DAY


class TimeDimensionSpec(DimensionSpec):  # noqa: D
    time_granularity: TimeGranularity = DEFAULT_TIME_GRANULARITY

    def column_associations(self, resolver: ColumnAssociationResolver) -> Tuple[ColumnAssociation, ...]:  # noqa: D
        return (resolver.resolve_time_dimension_spec(self),)

    def without_first_identifier_link(self) -> LinkableInstanceSpec:  # noqa: D
        assert len(self.identifier_links) > 0, f"Spec does not have any identifier links: {self}"
        return TimeDimensionSpec(
            element_name=self.element_name,
            identifier_links=self.identifier_links[1:],
            time_granularity=self.time_granularity,
        )

    def without_identifier_links(self) -> LinklessIdentifierSpec:  # noqa: D
        return LinklessIdentifierSpec.from_element_name(self.element_name)

    @staticmethod
    def from_name(name: str) -> TimeDimensionSpec:  # noqa: D
        structured_name = StructuredLinkableSpecName.from_name(name)
        return TimeDimensionSpec(
            identifier_links=tuple(
                LinklessIdentifierSpec.from_element_name(idl) for idl in structured_name.identifier_link_names
            ),
            element_name=structured_name.element_name,
            time_granularity=structured_name.time_granularity or DEFAULT_TIME_GRANULARITY,
        )

    @property
    def reference(self) -> TimeDimensionReference:  # noqa: D
        return TimeDimensionReference(element_name=self.element_name)

    @property
    def dimension_reference(self) -> DimensionReference:  # noqa: D
        return DimensionReference(element_name=self.element_name)

    @property
    def qualified_name(self) -> str:  # noqa: D
        return StructuredLinkableSpecName(
            identifier_link_names=tuple(x.element_name for x in self.identifier_links),
            element_name=self.element_name,
            time_granularity=self.time_granularity,
        ).qualified_name


@dataclass(frozen=True)
class ElementReference:
    """Used when we need to refer to a dimension, measure, identifier, but other attributes are unknown."""

    element_name: str


@dataclass(frozen=True)
class LinkableElementReference(ElementReference):
    """Used when we need to refer to a dimension or identifier, but other attributes are unknown."""

    pass


@dataclass(frozen=True)
class MeasureReference(ElementReference):
    """Used when we need to refer to a measure (separate from LinkableElementReference because measures aren't linkable"""

    pass


@dataclass(frozen=True)
class DimensionReference(LinkableElementReference):  # noqa: D
    pass

    @property
    def time_dimension_reference(self) -> TimeDimensionReference:  # noqa: D
        return TimeDimensionReference(element_name=self.element_name)


@dataclass(frozen=True)
class IdentifierReference(LinkableElementReference):  # noqa: D
    pass


@dataclass(frozen=True)
class CompositeSubIdentifierReference(ElementReference):  # noqa: D
    pass


@dataclass(frozen=True)
class TimeDimensionReference(DimensionReference):  # noqa: D
    pass

    def dimension_reference(self) -> DimensionReference:  # noqa: D
        return DimensionReference(element_name=self.element_name)


class MeasureSpec(InstanceSpec):  # noqa: D
    def column_associations(self, resolver: ColumnAssociationResolver) -> Tuple[ColumnAssociation, ...]:  # noqa: D
        return (resolver.resolve_measure_spec(self),)

    @staticmethod
    def from_name(name: str) -> MeasureSpec:
        """Construct from a name e.g. listing__ds__month."""
        return MeasureSpec(element_name=name)

    @property
    def qualified_name(self) -> str:  # noqa: D
        return self.element_name

    @property
    def as_reference(self) -> MeasureReference:  # noqa: D
        return MeasureReference(element_name=self.element_name)


class MetricSpec(InstanceSpec):  # noqa: D
    # Time-over-time could go here

    def column_associations(self, resolver: ColumnAssociationResolver) -> Tuple[ColumnAssociation, ...]:  # noqa: D
        return (resolver.resolve_metric_spec(self),)

    @property
    def qualified_name(self) -> str:  # noqa: D
        return self.element_name


class OrderBySpec(FrozenBaseModel):  # noqa: D
    item: InstanceSpec
    descending: bool


class FilterSpec(FrozenBaseModel):  # noqa: D
    expr: str
    elements: Tuple[InstanceSpec, ...]


class OutputColumnNameOverride(FrozenBaseModel):
    """Describes how we should name the output column for a time dimension instead of the default.

    Note: This is used temporarily to maintain compatibility with the old framework.
    """

    time_dimension_spec: TimeDimensionSpec
    output_column_name: str


class LinkableSpecSet(FrozenBaseModel):
    """Groups linkable specs."""

    dimension_specs: Tuple[DimensionSpec, ...] = ()
    time_dimension_specs: Tuple[TimeDimensionSpec, ...] = ()
    identifier_specs: Tuple[IdentifierSpec, ...] = ()

    @property
    def as_tuple(self) -> Tuple[LinkableInstanceSpec, ...]:  # noqa: D
        return tuple(itertools.chain(self.dimension_specs, self.time_dimension_specs, self.identifier_specs))

    @staticmethod
    def merge(spec_sets: Sequence[LinkableSpecSet]) -> LinkableSpecSet:
        """Merges and dedupes the linkable specs."""

        dimension_specs: List[DimensionSpec] = []
        time_dimension_specs: List[TimeDimensionSpec] = []
        identifier_specs: List[IdentifierSpec] = []

        for spec_set in spec_sets:
            for dimension_spec in spec_set.dimension_specs:
                if dimension_spec not in dimension_specs:
                    dimension_specs.append(dimension_spec)
            for time_dimension_spec in spec_set.time_dimension_specs:
                if time_dimension_spec not in time_dimension_specs:
                    time_dimension_specs.append(time_dimension_spec)
            for identifier_spec in spec_set.identifier_specs:
                if identifier_spec not in identifier_specs:
                    identifier_specs.append(identifier_spec)

        return LinkableSpecSet(
            dimension_specs=tuple(dimension_specs),
            time_dimension_specs=tuple(time_dimension_specs),
            identifier_specs=tuple(identifier_specs),
        )

    def is_subset_of(self, other_set: LinkableSpecSet) -> bool:  # noqa: D
        return set(self.as_tuple).issubset(set(other_set.as_tuple))


class SpecWhereClauseConstraint(FrozenBaseModel):
    """Similar to a WhereClauseConstraint, but with specs instead of strings"""

    # e.g. "listing__capacity_latest > 4"
    where_condition: str
    # e.g. {DimensionSpec(element_name="capacity_latest", identifier_links=("listing",))
    linkable_names: Tuple[str, ...]
    linkable_spec_set: LinkableSpecSet
    execution_parameters: SqlBindParameters

    def combine(self, other: SpecWhereClauseConstraint) -> SpecWhereClauseConstraint:  # noqa: D
        linkable_names = list(set(self.linkable_names).union(set(other.linkable_names)))
        where_condition = f"({self.where_condition}) AND ({other.where_condition})"
        new_sql_values = OrderedDict(self.execution_parameters.param_dict)
        for k, v in other.execution_parameters.param_dict.items():
            if k in new_sql_values and v != new_sql_values[k]:
                raise ValueError(
                    f"Cannot combine with an execution parameter collision. Both where clauses have key ({k}),"
                    f" but different values: ({v} != {new_sql_values[k]})"
                )

            new_sql_values[k] = v

        return SpecWhereClauseConstraint(
            where_condition=where_condition,
            linkable_names=tuple(linkable_names),
            linkable_spec_set=LinkableSpecSet.merge([self.linkable_spec_set, other.linkable_spec_set]),
            execution_parameters=SqlBindParameters(param_dict=OrderedDict(new_sql_values)),
        )


class MetricFlowQuerySpec(FrozenBaseModel):
    """Specs needed for running a query."""

    metric_specs: Tuple[MetricSpec, ...] = ()
    dimension_specs: Tuple[DimensionSpec, ...] = ()
    identifier_specs: Tuple[IdentifierSpec, ...] = ()
    time_dimension_specs: Tuple[TimeDimensionSpec, ...] = ()
    order_by_specs: Tuple[OrderBySpec, ...] = ()
    output_column_name_overrides: Tuple[OutputColumnNameOverride, ...] = ()
    time_range_constraint: Optional[TimeRangeConstraint] = None
    where_constraint: Optional[SpecWhereClauseConstraint] = None
    limit: Optional[int] = None

    @property
    def linkable_specs(self) -> LinkableSpecSet:  # noqa: D
        return LinkableSpecSet(
            dimension_specs=self.dimension_specs,
            time_dimension_specs=self.time_dimension_specs,
            identifier_specs=self.identifier_specs,
        )


TransformOutputT = TypeVar("TransformOutputT")


class InstanceSpecSetTransform(Generic[TransformOutputT], ABC):
    """Function to use for transforming spec sets."""

    @abstractmethod
    def transform(self, spec_set: InstanceSpecSet) -> TransformOutputT:  # noqa: D
        pass


class InstanceSpecSet(FrozenBaseModel):
    """Consolidates all specs used in an instance set"""

    metric_specs: Tuple[MetricSpec, ...] = ()
    measure_specs: Tuple[MeasureSpec, ...] = ()
    dimension_specs: Tuple[DimensionSpec, ...] = ()
    identifier_specs: Tuple[IdentifierSpec, ...] = ()
    time_dimension_specs: Tuple[TimeDimensionSpec, ...] = ()

    def merge(self, others: Sequence[InstanceSpecSet]) -> InstanceSpecSet:
        """Merge all sets into one set, without de-duplication."""
        return InstanceSpecSet(
            metric_specs=self.metric_specs + tuple(itertools.chain.from_iterable([x.metric_specs for x in others])),
            dimension_specs=self.dimension_specs
            + tuple(itertools.chain.from_iterable([x.dimension_specs for x in others])),
            identifier_specs=self.identifier_specs
            + tuple(itertools.chain.from_iterable([x.identifier_specs for x in others])),
            time_dimension_specs=self.time_dimension_specs
            + tuple(itertools.chain.from_iterable([x.time_dimension_specs for x in others])),
        )

    @property
    def linkable_specs(self) -> Sequence[LinkableInstanceSpec]:
        """All linkable specs in this set."""
        return list(itertools.chain(self.dimension_specs, self.time_dimension_specs, self.identifier_specs))

    @property
    def all_specs(self) -> Sequence[InstanceSpec]:  # noqa: D
        return tuple(
            itertools.chain(
                self.measure_specs,
                self.dimension_specs,
                self.time_dimension_specs,
                self.identifier_specs,
                self.metric_specs,
            )
        )

    def transform(self, transform_function: InstanceSpecSetTransform[TransformOutputT]) -> TransformOutputT:  # noqa: D
        return transform_function.transform(self)


class PartitionSpecSet(FrozenBaseModel):
    """Grouping of the linkable specs."""

    dimension_specs: Tuple[DimensionSpec, ...] = ()
    time_dimension_specs: Tuple[TimeDimensionSpec, ...] = ()



from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar, Dict

from metricflow.model.objects.data_source import DataSource

T = TypeVar("T")


class DataSourceContainer(ABC, Generic[T]):  # noqa: D
    @abstractmethod
    def get(self, data_source_name: str) -> T:  # noqa: D
        pass

    @abstractmethod
    def values(self) -> List[T]:  # noqa: D
        pass

    @abstractmethod
    def keys(self) -> List[str]:  # noqa: D
        pass

    @abstractmethod
    def __contains__(self, item: str) -> bool:  # noqa: D
        pass

    @abstractmethod
    def put(self, key: str, value: T) -> None:  # noqa: D
        pass


class PydanticDataSourceContainer(DataSourceContainer[DataSource]):  # noqa: D
    def __init__(self, data_sources: List[DataSource]) -> None:  # noqa: D
        self._data_source_index: Dict[str, DataSource] = {data_source.name: data_source for data_source in data_sources}

    def get(self, data_source_name: str) -> DataSource:  # noqa: D
        return self._data_source_index[data_source_name]

    def values(self) -> List[DataSource]:  # noqa: D
        return list(self._data_source_index.values())

    def put(self, key: str, value: T) -> None:  # noqa: D
        raise TypeError("Cannot call put on static data source container. Data sources are fixed on init")

    def _put(self, data_source: DataSource) -> None:
        """Dont use this unless you mean it (ie in tests). This is supposed to be static"""
        self._data_source_index[data_source.name] = data_source

    def keys(self) -> List[str]:  # noqa: D
        return list(self._data_source_index.keys())

    def __contains__(self, item: str) -> bool:  # noqa: D
        return item in self._data_source_index
